# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
# Copyright 2016 FUJITSU LIMITED

"""Classes for monitoring the monitoring server stack.

    Covering mon-persister, mon-api and mon-thresh.
    Kafka, mysql, vertica and influxdb are covered by other detection plugins. Mon-notification uses statsd.
"""

import logging
import yaml

from six.moves import configparser

import monasca_setup.agent_config
import monasca_setup.detection
from monasca_setup.detection import find_process_cmdline
from monasca_setup.detection import find_process_name
from monasca_setup.detection.utils import watch_process
from monasca_setup.detection.utils import watch_process_by_username

log = logging.getLogger(__name__)

_PYTHON_LANG_MARKERS = 'python', 'gunicorn',
"""List of all strings that if found in process exe
mean that application runs under Python"""
_JAVA_LANG_MARKERS = 'java',

_DEFAULT_API_PORT = 8070
"""Default TCP port which monasca-api process should be available by"""


def _get_impl_lang(process):
    """Return implementation language of the application behind process

    :param process: current process
    :type process: psutil.Process
    :return: implementation lang, either java or python
    :rtype: str

    """
    p_exe = process.as_dict()['exe']
    for lm in _PYTHON_LANG_MARKERS:
        if lm in p_exe:
            return 'python'
    for lm in _JAVA_LANG_MARKERS:
        if lm in p_exe:
            return 'java'
    raise Exception(('Cannot determine language '
                     'implementation from process exe %s') % p_exe)


class MonAgent(monasca_setup.detection.Plugin):
    """Detect the Monsaca agent engine and setup some simple checks."""
    def _detect(self):
        """Run detection, set self.available True if the service is detected."""
        self.available = True
        agent_process_list = ['monasca-collector', 'monasca-forwarder',
                              'monasca-statsd']
        for process in agent_process_list:
            if find_process_cmdline(process) is None:
                self.available = False
                return

    def build_config(self):
        """Build the config as a Plugins object and return."""
        log.info("\tEnabling the Monasca Agent process check")
        return watch_process_by_username('mon-agent', 'monasca-agent', 'monitoring',
                                         'monasca-agent')

    def dependencies_installed(self):
        return True


class MonAPI(monasca_setup.detection.Plugin):
    """Detect mon_api and setup monitoring."""

    PARTIAL_ERR_MSG = 'Plugin for monasca-api will not be configured.'

    def _detect(self):
        """Detects if monasca-api runs in the system

        Method distinguishes between Java and Python implementation
        hence provides different agent's configuration.

        """

        def check_port():
            for conn in api_process.connections('inet'):
                if conn.laddr[1] == api_port:
                    return True
            return False

        api_process = find_process_cmdline('monasca-api')
        process_found = api_process is not None

        if process_found:
            impl_lang = _get_impl_lang(api_process)
            impl_helper = self._init_impl_helper(impl_lang)

            impl_helper.load_configuration()

            api_port = impl_helper.get_bound_port()
            port_taken = check_port()

            if not port_taken:
                log.error('monasca-api is not listening on port %d. %s'
                          % (api_port, self.PARTIAL_ERR_MSG))
                return

            log.info('\tmonasca-api implementation is %s', impl_lang)

            self.available = port_taken
            self._impl_helper = impl_helper
        else:
            log.error('monasca-api process has not been found. %s'
                      % self.PARTIAL_ERR_MSG)

    def build_config(self):
        """Build the config as a Plugins object and return."""
        config = monasca_setup.agent_config.Plugins()

        log.info("\tEnabling the monasca-api process check")
        config.merge(watch_process(
            search_strings=['monasca-api'],
            service='monitoring',
            component='monasca-api',
            exact_match=False
        ))
        impl_config = self._impl_helper.build_config()
        if impl_config:
            config.merge(impl_config)

        return config

    def dependencies_installed(self):
        return True

    @staticmethod
    def _init_impl_helper(impl_lang):
        """Returns appropriate helper implementation.

        :param impl_lang: implementation language, either `java` or `python`
        :type impl_lang: str
        :return: implementation helper
        :rtype: Union(_MonAPIJavaHelper,_MonAPIPythonHelper)

        """
        if impl_lang == 'java':
            return _MonAPIJavaHelper()
        else:
            return _MonAPIPythonHelper()


class MonNotification(monasca_setup.detection.Plugin):
    """Detect the Monsaca notification engine and setup some simple checks."""
    def _detect(self):
        """Run detection, set self.available True if the service is detected."""
        if find_process_cmdline('monasca-notification') is not None:
            self.available = True

    def build_config(self):
        """Build the config as a Plugins object and return."""
        log.info("\tEnabling the Monasca Notification healthcheck")
        return watch_process_by_username('mon-notification', 'monasca-notification', 'monitoring',
                                         'monasca-notification')

    def dependencies_installed(self):
        return True


class MonPersister(monasca_setup.detection.Plugin):
    """Detect mon_persister and setup monitoring."""

    PARTIAL_ERR_MSG = 'Plugin for monasca-persister will not be configured.'

    def _detect(self):
        """Detects if monasca-persister runs in the system

        Method distinguishes between Java and Python implementation
        hence provides different agent's configuration.

        """
        p_process = find_process_cmdline('monasca-persister')
        process_found = p_process is not None

        if process_found:
            impl_lang = _get_impl_lang(p_process)
            impl_helper = self._init_impl_helper(impl_lang)

            if impl_helper is not None:
                impl_helper.load_configuration()

            self._impl_helper = impl_helper
            self.available = True

            log.info('\tmonasca-persister implementation is %s', impl_lang)

        else:
            log.error('monasca-persister process has not been found. %s'
                      % self.PARTIAL_ERR_MSG)

    def build_config(self):
        """Build the config as a Plugins object and return."""
        config = monasca_setup.agent_config.Plugins()

        log.info("\tEnabling the Monasca persister process check")
        config.merge(watch_process(
            search_strings=['monasca-persister'],
            service='monitoring',
            component='monasca-persister',
            exact_match=False
        ))
        if self._impl_helper is not None:
            impl_config = self._impl_helper.build_config()
            if impl_config:
                config.merge(impl_config)

        return config

    def dependencies_installed(self):
        return True

    @staticmethod
    def _init_impl_helper(impl_lang):
        """Returns appropriate helper implementation.

        Note:

            This method returns the helper only for Java.
            Python implementation comes with no extra setup.

        :param impl_lang: implementation language, either `java` or `python`
        :type impl_lang: str
        :return: implementation helper
        :rtype:_MonPersisterJavaHelper

        """
        if impl_lang == 'java':
            return _MonPersisterJavaHelper()
        return None


class MonThresh(monasca_setup.detection.Plugin):
    """Detect the running mon-thresh and monitor."""
    def _detect(self):
        """Run detection, set self.available True if the service is detected."""
        # The node will be running either nimbus or supervisor or both
        self.available = (find_process_cmdline('storm.daemon.nimbus') is not None or
                          find_process_cmdline('storm.daemon.supervisor') is not None)

    def build_config(self):
        """Build the config as a Plugins object and return."""
        log.info("\tWatching the mon-thresh process.")
        config = monasca_setup.agent_config.Plugins()
        for process in ['storm.daemon.nimbus', 'storm.daemon.supervisor', 'storm.daemon.worker']:
            if find_process_cmdline(process) is not None:
                config.merge(watch_process([process], 'monitoring', 'apache-storm', exact_match=False, detailed=False))
        config.merge(watch_process_by_username('storm', 'monasca-thresh', 'monitoring', 'apache-storm'))
        return config

    def dependencies_installed(self):
        return True


def dropwizard_health_check(service, component, url):
    """Setup a dropwizard heathcheck to be watched by the http_check plugin."""
    config = monasca_setup.agent_config.Plugins()
    config['http_check'] = {'init_config': None,
                            'instances': [{'name': "{0}-{1} healthcheck".format(service, component),
                                           'url': url,
                                           'timeout': 5,
                                           'include_content': False,
                                           'dimensions': {'service': service, 'component': component}}]}
    return config


def dropwizard_metrics(service, component, url, whitelist):
    """Setup a dropwizard metrics check"""
    config = monasca_setup.agent_config.Plugins()
    config['http_metrics'] = {'init_config': None,
                              'instances': [{'name': "{0}-{1} metrics".format(service, component),
                                             'url': url,
                                             'timeout': 5,
                                             'dimensions': {'service': service, 'component': component},
                                             'whitelist': whitelist}]}
    return config


class MonInfluxDB(monasca_setup.detection.Plugin):
    """Detect InfluxDB and setup some simple checks."""

    def _detect(self):
        """Run detection, set self.available True if the service is detected.
        """

        if find_process_name('influxd') is not None:
            self.available = True

    def build_config(self):
        """Build the config as a Plugins object and return."""
        log.info("\tEnabling the Monasca InfluxDB check")
        return watch_process(['influxd'], 'monitoring', 'influxd',
                             exact_match=False)

    def dependencies_installed(self):
        return True


class _MonPersisterJavaHelper(object):
    """Encapsulates Java specific configuration for monasca-persister"""

    CONFIG_FILE = '/etc/monasca/persister-config.yml'
    """Default location where plugin expects configuration file"""

    def __init__(self):
        super(_MonPersisterJavaHelper, self).__init__()
        self._cfg = None

    def build_config(self):
        config = monasca_setup.agent_config.Plugins()
        metrics = self._collect_metrics()
        self._monitor_endpoints(config, metrics)
        return config

    def _collect_metrics(self):
        """Collects all the metrics .

        Methods will return list of all metrics that will
        later be used in
        :py:mod:`monasca_agent.collector.checks_d.http_metrics` to query
        admin endpoint of monasca-persister.

        Following group of metrics are examined:

        * JVM metrics
        * DB metrics ( see also :py:meth:`._collect_db_metrics` )
        * Internal metrics ( see also :py:meth:`._collect_internal_metrics` )

        :return: list of metrics
        :rtype: list

        """
        log.info("\tEnabling the monasca-persister metrics")

        whitelist = [
            {
                "name": "jvm.memory.total.max",
                "path": "gauges/jvm.memory.total.max/value",
                "type": "gauge"
            },
            {
                "name": "jvm.memory.total.used",
                "path": "gauges/jvm.memory.total.used/value",
                "type": "gauge"
            }
        ]

        self._collect_db_metrics(whitelist)
        self._collect_internal_metrics(whitelist)

        return whitelist

    def _collect_db_metrics(self, whitelist):
        """Collects DB specific metrics.

        Method retrieves which time-series database is used
        in monaca-persister and sets up new metrics to be monitored.

        Note:
            Only if vertica is TSDB in monasca-persister

        """
        database_type = self._cfg['databaseConfiguration']['databaseType']
        if database_type == 'influxdb':
            pass
        elif database_type == 'vertica':
            self._add_vertica_metrics(whitelist)
        else:
            log.warn('Failed finding database type in %s', self.CONFIG_FILE)

    def _collect_internal_metrics(self, whitelist):
        alarm_num_threads = self._cfg['alarmHistoryConfiguration']['numThreads']
        metric_num_threads = self._cfg['metricConfiguration']['numThreads']

        # Dynamic Whitelist
        for idx in range(alarm_num_threads):
            new_thread = {
                "name": "alarm-state-transitions-added-to-batch-counter[{0}]".format(idx),
                "path": "counters/monasca.persister.pipeline.event.AlarmStateTransitionHandler[alarm-state-transition-{0}].alarm-state-transitions-added-to-batch-counter/count".format(idx),
                "type": "rate"
            }
            whitelist.append(new_thread)
        for idx in range(metric_num_threads):
            new_thread = {
                "name": "metrics-added-to-batch-counter[{0}]".format(idx),
                "path": "counters/monasca.persister.pipeline.event.MetricHandler[metric-{0}].metrics-added-to-batch-counter/count".format(idx),
                "type": "rate"
            }
            whitelist.append(new_thread)

    def _add_vertica_metrics(self, whitelist):
        whitelist.extend([
            {
                "name": "monasca.persister.repository.vertica.VerticaMetricRepo.definition-cache-hit-meter",
                "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo.definition-cache-hit-meter/count",
                "type": "rate"
            },
            {
                "name": "monasca.persister.repository.vertica.VerticaMetricRepo.definition-cache-miss-meter",
                "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo.definition-cache-miss-meter/count",
                "type": "rate"
            },
            {
                "name": "monasca.persister.repository.vertica.VerticaMetricRepo.definition-dimension-cache-hit-meter",
                "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo.definition-dimension-cache-hit-meter/count",
                "type": "rate"
            },
            {
                "name": "monasca.persister.repository.vertica.VerticaMetricRepo.definition-dimension-cache-miss-meter",
                "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo.definition-dimension-cache-miss-meter/count",
                "type": "rate"
            },
            {
                "name": "monasca.persister.repository.vertica.VerticaMetricRepo.dimension-cache-hit-meter",
                "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo.dimension-cache-hit-meter/count",
                "type": "rate"
            },
            {
                "name": "monasca.persister.repository.vertica.VerticaMetricRepo.dimension-cache-miss-meter",
                "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo.dimension-cache-miss-meter/count",
                "type": "rate"
            },
            {
                "name": "monasca.persister.repository.vertica.VerticaMetricRepo.measurement-meter",
                "path": "meters/monasca.persister.repository.vertica.VerticaMetricRepo.measurement-meter/count",
                "type": "rate"
            }
        ])

    def _monitor_endpoints(self, config, metrics):
        admin_connector = self._cfg['server']['adminConnectors'][0]

        try:
            admin_endpoint_type = admin_connector['type']
        except Exception:
            admin_endpoint_type = "http"
        try:
            admin_endpoint_port = admin_connector['port']
        except Exception:
            admin_endpoint_port = 8091

        log.info("\tEnabling the Monasca persister healthcheck")
        config.merge(
            dropwizard_health_check(
                'monitoring',
                'monasca-persister',
                '{0}://localhost:{1}/healthcheck'.format(admin_endpoint_type,
                                                         admin_endpoint_port)))

        log.info("\tEnabling the Monasca persister metrics")
        config.merge(
            dropwizard_metrics(
                'monitoring',
                'monasca-persister',
                '{0}://localhost:{1}/metrics'.format(admin_endpoint_type,
                                                     admin_endpoint_port),
                metrics))

    def load_configuration(self):
        """Loads java specific configuration.

        Load java specific configuration from:

        * :py:attr:`API_CONFIG_YML`

        :return: True if both configuration files were successfully loaded
        :rtype: bool

        """
        try:
            with open(self.CONFIG_FILE, 'r') as config:
                self._cfg = yaml.safe_load(config.read())
        except Exception as ex:
            log.error('Failed to parse %s', self.CONFIG_FILE)
            log.exception(ex)
            raise ex


class _MonAPIJavaHelper(object):
    """Encapsulates Java specific configuration for monasca-api"""

    CONFIG_FILE = '/etc/monasca/api-config.yml'

    def __init__(self):
        super(_MonAPIJavaHelper, self).__init__()
        self._api_config = None

    def build_config(self):
        """Builds monitoring configuration for monasca-api Java flavour.

        Method configures additional checks that are specific for
        Java implementation of monasca-api.

        List of checks:

        * HttpCheck, :py:mod:`monasca_agent.collector.checks_d.http_check`
        * HttpMetrics, :py:mod:`monasca_agent.collector.checks_d.http_metrics`

        """
        config = monasca_setup.agent_config.Plugins()

        log.info("\tEnabling the Monasca api metrics")
        whitelist = [
            {
                "name": "jvm.memory.total.max",
                "path": "gauges/jvm.memory.total.max/value",
                "type": "gauge"
            },
            {
                "name": "jvm.memory.total.used",
                "path": "gauges/jvm.memory.total.used/value",
                "type": "gauge"
            },
            {
                "name": "metrics.published",
                "path": "meters/monasca.api.app.MetricService.metrics.published/count",
                "type": "rate"
            }
        ]

        if not self._is_hibernate_on():
            # if hibernate is not used, it is mysql with DBI
            # for that case having below entries makes sense
            log.debug(
                'monasca-api has not enabled Hibernate, adding DBI metrics')
            whitelist.extend([
                {
                    "name": "raw-sql.time.avg",
                    "path": "timers/org.skife.jdbi.v2.DBI.raw-sql/mean",
                    "type": "gauge"
                },
                {
                    "name": "raw-sql.time.max",
                    "path": "timers/org.skife.jdbi.v2.DBI.raw-sql/max",
                    "type": "gauge"
                }
            ])

        self._monitor_endpoints(config, whitelist)

        return config

    def _monitor_endpoints(self, config, metrics):
        admin_connector = self._api_config['server']['adminConnectors'][0]

        try:
            admin_endpoint_type = admin_connector['type']
        except Exception:
            admin_endpoint_type = "http"
        try:
            admin_endpoint_port = admin_connector['port']
        except Exception:
            admin_endpoint_port = 8081

        healthcheck_url = '{0}://localhost:{1}/healthcheck'.format(
            admin_endpoint_type, admin_endpoint_port)
        metric_url = '{0}://localhost:{1}/metrics'.format(
            admin_endpoint_type, admin_endpoint_port)

        log.info("\tEnabling the monasca-api healthcheck")
        config.merge(dropwizard_health_check('monitoring', 'monasca-api',
                                             healthcheck_url))
        log.info("\tEnabling the monasca-api metrics")
        config.merge(dropwizard_metrics('monitoring', 'monasca-api',
                                        metric_url, metrics))

    def _is_hibernate_on(self):
        # check if api_config has been declared in __init__
        # if not it means that configuration file was not found
        # or monasca-api Python implementation is running

        api_config = getattr(self, '_api_config', None)
        if api_config is None:
            return False

        hibernate_cfg = self._api_config.get('hibernate', None)
        if hibernate_cfg is None:
            return False

        return hibernate_cfg.get('supportEnabled', False)

    def load_configuration(self):
        """Loads java specific configuration.

        Load java specific configuration from:

        * :py:attr:`API_CONFIG_YML`

        :return: True if both configuration files were successfully loaded
        :rtype: bool

        """
        try:
            with open(self.CONFIG_FILE, 'r') as config:
                self._api_config = yaml.safe_load(config.read())
        except Exception as ex:
            log.error('Failed to parse %s', self.CONFIG_FILE)
            log.exception(ex)
            raise ex

    def get_bound_port(self):
        """Returns port API is listening on.

        Method tries to read port from the '/etc/monasca/api-config.yml'
        file. In case if:

        * file was not found in specified location
        * file could be read from the file system
        * file was changed and assumed location of port changed

        code rollbacks to :py:const:`_DEFAULT_API_PORT`

        :return: TCP port api is listening on
        :rtype: int

        """
        if self._api_config is None:
            return _DEFAULT_API_PORT
        try:
            return self._api_config['server']['applicationConnectors'][0]['port']
        except Exception as ex:
            log.error('Failed to read api port from '
                      '/etc/monasca/api-config.yml')
            log.exception(ex)
            return _DEFAULT_API_PORT


class _MonAPIPythonHelper(object):
    """Encapsulates Python specific configuration for monasca-api"""

    CONFIG_FILE = '/etc/monasca/api-config.ini'

    def __init__(self):
        super(_MonAPIPythonHelper, self).__init__()
        self._paste_config = None

    def build_config(self):
        # note(trebskit) intentionally left empty because gunicorn check
        # seems to have some serious issues with monitoring gunicorn process
        # see https://bugs.launchpad.net/monasca/+bug/1646481
        return None

    def load_configuration(self):
        """Loads INI file from specified path.

        Method loads configuration from specified `path`
        and parses it with :py:class:`configparser.RawConfigParser`

        """
        cp = configparser.RawConfigParser()
        try:
            cp.readfp(open(self.CONFIG_FILE, 'r'))
            self._paste_config = cp
        except Exception as ex:
            log.error('Failed to parse %s', self.CONFIG_FILE)
            log.exception(ex)
            raise ex

    def get_bound_port(self):
        """Returns port API is listening on.

        Method tries to read port from the '/etc/monasca/api-config.ini'
        file. In case if:

        * file was not found in specified location
        * file could be read from the file system
        * file was changed and assumed location of port changed

        code rollbacks to :py:const:`_DEFAULT_API_PORT`

        :return: TCP port api is listening on
        :rtype: int

        """
        if not self._paste_config:
            return _DEFAULT_API_PORT
        return self._paste_config.getint('server:main', 'port')



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging
import os

import monasca_setup.agent_config
import monasca_setup.detection


log = logging.getLogger(__name__)

# Defaults
supervisord_conf = '/root/.supervisord.cnf'
supervisord_server_name = 'server0'


class Supervisord(monasca_setup.detection.Plugin):
    """Detect supervisord process and setup configuration for monitoring.

        This plugin needs connection info for supervisord setup. There are two
        ways to provide it, either by a file placed in /root/.supervisord.cnf
        or by specifying the following arguments:
            - server (req, arbitrary name to identify the supervisord server)
            - socket (opt, required for socket connection type)
            - host (opt, defaults to localhost)
            - port (opt, defaults to 9001)
            - user (opt, only if username is configured)
            - password (opt, only if password is configured)
            - process_regex (opt, regex patterns for processes to monitor)
            - process_names (opt, process to monitor by name)
        process_regex and process_names are comma separated lists

        The file at /root/.supervisord.cnf should have this format:
        [client]
            server=server0
            socket=unix:///var/run//supervisor.sock
            process_names=apache2,webapp,java
    """

    def _detect(self):
        """Run detection, set self.available True if the service is detected.
        """
        found_process = (monasca_setup.detection.
                         find_process_cmdline('supervisord'))
        has_args_or_conf_file = (self.args is not None or
                                 os.path.isfile(supervisord_conf))
        self.available = found_process is not None and has_args_or_conf_file
        if not self.available:
            if not found_process:
                log.error('Supervisord process does not exist.')
            elif not has_args_or_conf_file:
                log.error(('Supervisord process exists but '
                           'configuration file was not found and '
                           'no arguments were given.'))

    def _get_config(self):
        """Set the configuration to be used for connecting to supervisord
        :return:
        """
        # Set defaults and read config or use arguments
        if self.args is None:
            self.server = supervisord_server_name
            self.socket = None
            self.host = None
            self.port = None
            self.user = None
            self.password = None
            self.process_regex = None
            self.process_names = None
            self.process_details_check = None
            self.process_uptime_check = None

            self._read_config(supervisord_conf)
        else:
            self.server = self.args.get('server', supervisord_server_name)
            self.socket = self.args.get('socket')
            self.host = self.args.get('host')
            self.port = self.args.get('port')
            self.user = self.args.get('user')
            self.password = self.args.get('pass')
            self.process_regex = self.args.get('proc_regex')
            self.process_names = self.args.get('proc_names')
            self.process_details_check = self.args.get('proc_details_check')
            self.process_uptime_check = self.args.get('proc_uptime_check')

    def _read_config(self, config_file):
        """Read the configuration setting member variables as appropriate.
        :param config_file: The filename of the configuration to read and parse
        """
        # Read the supervisord config file to extract the needed variables.
        client_section = False
        try:
            with open(config_file, "r") as conf:
                for row in conf:
                    if "[client]" in row:
                        client_section = True
                        log.info("\tUsing client credentials from {:s}".format(config_file))
                        pass
                    if client_section:
                        if "server=" in row:
                            self.server = row.split("=")[1].strip()
                        if "socket=" in row:
                            self.socket = row.split("=")[1].strip()
                        if "host=" in row:
                            self.host = row.split("=")[1].strip()
                        if "port=" in row:
                            self.port = row.split("=")[1].strip()
                        if "user=" in row:
                            self.user = row.split("=")[1].strip()
                        if "pass=" in row:
                            self.password = row.split("=")[1].strip()
                        if "proc_regex=" in row:
                            self.process_regex = row.split("=")[1].strip()
                        if "proc_names=" in row:
                            self.process_names = row.split("=")[1].strip()
                        if "proc_details_check=" in row:
                            self.process_details_check = row.split("=")[1].strip()
                        if "proc_uptime_check=" in row:
                            self.process_uptime_check = row.split("=")[1].strip()
        except IOError:
            log.error("\tI/O error reading {:s}".format(config_file))

    @staticmethod
    def _split_list(to_split):
        return [x.strip() for x in to_split.split(',')]

    def build_config(self):
        """Build the config as a Plugins object and return.
        """
        config = monasca_setup.agent_config.Plugins()
        # First watch the process
        config.merge(monasca_setup.detection.watch_process(['supervisord'],
                                                           'supervisord',
                                                           exact_match=False))
        log.info("\tWatching the supervisord process.")

        try:
            self._get_config()
            instance_config = {'name': self.server}
            if self.socket is not None:
                instance_config['socket'] = self.socket
            if self.host is not None:
                instance_config['host'] = self.host
            if self.port is not None:
                instance_config['port'] = self.port
            if self.user is not None:
                instance_config['user'] = self.user
            if self.password is not None:
                instance_config['pass'] = self.password
            if self.process_regex is not None:
                instance_config['proc_regex'] = self._split_list(self.process_regex)
            if self.process_names is not None:
                instance_config['proc_names'] = self._split_list(self.process_names)
            if self.process_details_check is not None:
                instance_config['proc_details_check'] = self.process_details_check
            if self.process_uptime_check is not None:
                instance_config['proc_uptime_check'] = self.process_uptime_check

            config['supervisord'] = {'init_config': None, 'instances': [instance_config]}
        except Exception:
            log.exception('Error configuring the supervisord check plugin')

        return config

    def dependencies_installed(self):
        return True



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
""" Metric data types
"""
import logging

log = logging.getLogger(__name__)


class Metric(object):
    """A base metric class """

    def __init__(self, name, dimensions, tenant):
        self.metric = {'name': name,
                       'dimensions': dimensions.copy()}

        self.value_meta = None
        self.value = None
        self.timestamp = None
        self.tenant = tenant

    def measurement(self, value, timestamp):
        measurement = self.metric.copy()

        if self.value_meta:
            measurement['value_meta'] = self.value_meta.copy()
        else:
            measurement['value_meta'] = None

        measurement['value'] = value
        measurement['timestamp'] = timestamp * 1000

        envelope = {'measurement': measurement,
                    'tenant_id': self.tenant}

        return envelope

    def sample(self, value, sample_rate, timestamp):
        """Save a sample. """
        raise NotImplementedError()

    def flush(self):
        """Flush current sample. """
        raise NotImplementedError()


class Gauge(Metric):
    """A metric that tracks a value at particular points in time. """

    def __init__(self, name, dimensions, tenant=None):
        super(Gauge, self).__init__(name, dimensions, tenant)

    def sample(self, value, sample_rate, timestamp):
        self.value = value
        self.timestamp = timestamp

    def flush(self):
        # 0 is a valid value, so can't do: if not self.value:
        if self.value is None:
            return []

        envelope = self.measurement(self.value, self.timestamp)
        self.value = None
        return [envelope]


class Counter(Metric):
    """A metric that tracks a counter value. """

    def __init__(self, name, dimensions, tenant=None):
        super(Counter, self).__init__(name, dimensions, tenant)
        self.value = 0

    def sample(self, value, sample_rate, timestamp):
        try:
            self.value += value * int(1 / sample_rate)
            self.timestamp = timestamp
        except TypeError:
            log.error("metric {} value {} sample_rate {}".
                      format(self.metric['name'], value, sample_rate))

    def flush(self):
        envelope = self.measurement(self.value, self.timestamp)
        self.value = 0
        return [envelope]


class Rate(Metric):
    """Track the rate of metrics over each flush interval """

    def __init__(self, name, dimensions, tenant=None):
        super(Rate, self).__init__(name, dimensions, tenant)
        self.samples = []

    def sample(self, value, sample_rate, timestamp):
        self.samples.append((int(timestamp), value))
        self.timestamp = timestamp

        if len(self.samples) < 2:
            self.value = None
        else:
            self.value = self._rate(self.samples[-2], self.samples[-1])
            self.samples = self.samples[-1:]

    def _rate(self, sample1, sample2):
        delta_t = sample2[0] - sample1[0]
        delta_v = sample2[1] - sample1[1]
        rate = None
        if delta_v < 0:
            log.debug('Metric {0} has a rate < 0. New value = {1} and old '
                      'value = {2}. Counter may have been Reset.'.
                      format(self.metric['name'], sample2[1], sample1[1]))
            return rate
        try:
            rate = delta_v / float(delta_t)
        except ZeroDivisionError as e:
            log.exception('Error in sampling metric {0}, time difference '
                          'between current time and last_update time is '
                          '0, returned {1}'.
                          format(self.metric['name'], e))
        return rate

    def flush(self):
        if self.value is None:
            return []

        envelope = self.measurement(self.value, self.timestamp)
        self.value = None
        return [envelope]



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development Company LP

import json
import socket
import subprocess
import sys
import urllib2
import urlparse

from monasca_agent.collector.checks import AgentCheck
from monasca_agent.collector.checks.utils import add_basic_auth
from monasca_agent.common.util import headers


class NodeNotFound(Exception):
    pass


class ElasticSearch(AgentCheck):
    METRICS = {  # Metrics that are common to all Elasticsearch versions
        "elasticsearch.docs.count": ("gauge", "indices.docs.count"),
        "elasticsearch.docs.deleted": ("gauge", "indices.docs.deleted"),
        "elasticsearch.store.size": ("gauge", "indices.store.size_in_bytes"),
        "elasticsearch.indexing.index.total": ("gauge", "indices.indexing.index_total"),
        "elasticsearch.indexing.index.time": ("gauge", "indices.indexing.index_time_in_millis", lambda v: float(v) / 1000),
        "elasticsearch.indexing.index.current": ("gauge", "indices.indexing.index_current"),
        "elasticsearch.indexing.delete.total": ("gauge", "indices.indexing.delete_total"),
        "elasticsearch.indexing.delete.time": ("gauge", "indices.indexing.delete_time_in_millis", lambda v: float(v) / 1000),
        "elasticsearch.indexing.delete.current": ("gauge", "indices.indexing.delete_current"),
        "elasticsearch.get.total": ("gauge", "indices.get.total"),
        "elasticsearch.get.time": ("gauge", "indices.get.time_in_millis", lambda v: float(v) / 1000),
        "elasticsearch.get.current": ("gauge", "indices.get.current"),
        "elasticsearch.get.exists.total": ("gauge", "indices.get.exists_total"),
        "elasticsearch.get.exists.time": ("gauge", "indices.get.exists_time_in_millis", lambda v: float(v) / 1000),
        "elasticsearch.get.missing.total": ("gauge", "indices.get.missing_total"),
        "elasticsearch.get.missing.time": ("gauge", "indices.get.missing_time_in_millis", lambda v: float(v) / 1000),
        "elasticsearch.search.query.total": ("gauge", "indices.search.query_total"),
        "elasticsearch.search.query.time": ("gauge", "indices.search.query_time_in_millis", lambda v: float(v) / 1000),
        "elasticsearch.search.query.current": ("gauge", "indices.search.query_current"),
        "elasticsearch.search.fetch.total": ("gauge", "indices.search.fetch_total"),
        "elasticsearch.search.fetch.time": ("gauge", "indices.search.fetch_time_in_millis", lambda v: float(v) / 1000),
        "elasticsearch.search.fetch.current": ("gauge", "indices.search.fetch_current"),
        "elasticsearch.merges.current": ("gauge", "indices.merges.current"),
        "elasticsearch.merges.current.docs": ("gauge", "indices.merges.current_docs"),
        "elasticsearch.merges.current.size": ("gauge", "indices.merges.current_size_in_bytes"),
        "elasticsearch.merges.total": ("gauge", "indices.merges.total"),
        "elasticsearch.merges.total.time": ("gauge", "indices.merges.total_time_in_millis", lambda v: float(v) / 1000),
        "elasticsearch.merges.total.docs": ("gauge", "indices.merges.total_docs"),
        "elasticsearch.merges.total.size": ("gauge", "indices.merges.total_size_in_bytes"),
        "elasticsearch.refresh.total": ("gauge", "indices.refresh.total"),
        "elasticsearch.refresh.total.time": ("gauge", "indices.refresh.total_time_in_millis", lambda v: float(v) / 1000),
        "elasticsearch.flush.total": ("gauge", "indices.flush.total"),
        "elasticsearch.flush.total.time": ("gauge", "indices.flush.total_time_in_millis", lambda v: float(v) / 1000),
        "elasticsearch.process.open_fd": ("gauge", "process.open_file_descriptors"),
        "elasticsearch.transport.rx_count": ("gauge", "transport.rx_count"),
        "elasticsearch.transport.tx_count": ("gauge", "transport.tx_count"),
        "elasticsearch.transport.rx_size": ("gauge", "transport.rx_size_in_bytes"),
        "elasticsearch.transport.tx_size": ("gauge", "transport.tx_size_in_bytes"),
        "elasticsearch.transport.server_open": ("gauge", "transport.server_open"),
        "elasticsearch.thread_pool.bulk.active": ("gauge", "thread_pool.bulk.active"),
        "elasticsearch.thread_pool.bulk.threads": ("gauge", "thread_pool.bulk.threads"),
        "elasticsearch.thread_pool.bulk.queue": ("gauge", "thread_pool.bulk.queue"),
        "elasticsearch.thread_pool.bulk.rejected": ("gauge", "thread_pool.bulk.rejected"),
        "elasticsearch.thread_pool.flush.active": ("gauge", "thread_pool.flush.active"),
        "elasticsearch.thread_pool.flush.threads": ("gauge", "thread_pool.flush.threads"),
        "elasticsearch.thread_pool.flush.queue": ("gauge", "thread_pool.flush.queue"),
        "elasticsearch.thread_pool.flush.rejected": ("gauge", "thread_pool.flush.rejected"),
        "elasticsearch.thread_pool.generic.active": ("gauge", "thread_pool.generic.active"),
        "elasticsearch.thread_pool.generic.threads": ("gauge", "thread_pool.generic.threads"),
        "elasticsearch.thread_pool.generic.queue": ("gauge", "thread_pool.generic.queue"),
        "elasticsearch.thread_pool.generic.rejected": ("gauge", "thread_pool.generic.rejected"),
        "elasticsearch.thread_pool.get.active": ("gauge", "thread_pool.get.active"),
        "elasticsearch.thread_pool.get.threads": ("gauge", "thread_pool.get.threads"),
        "elasticsearch.thread_pool.get.queue": ("gauge", "thread_pool.get.queue"),
        "elasticsearch.thread_pool.get.rejected": ("gauge", "thread_pool.get.rejected"),
        "elasticsearch.thread_pool.index.active": ("gauge", "thread_pool.index.active"),
        "elasticsearch.thread_pool.index.threads": ("gauge", "thread_pool.index.threads"),
        "elasticsearch.thread_pool.index.queue": ("gauge", "thread_pool.index.queue"),
        "elasticsearch.thread_pool.index.rejected": ("gauge", "thread_pool.index.rejected"),
        "elasticsearch.thread_pool.management.active": ("gauge", "thread_pool.management.active"),
        "elasticsearch.thread_pool.management.threads": ("gauge", "thread_pool.management.threads"),
        "elasticsearch.thread_pool.management.queue": ("gauge", "thread_pool.management.queue"),
        "elasticsearch.thread_pool.management.rejected": ("gauge", "thread_pool.management.rejected"),
        "elasticsearch.thread_pool.merge.active": ("gauge", "thread_pool.merge.active"),
        "elasticsearch.thread_pool.merge.threads": ("gauge", "thread_pool.merge.threads"),
        "elasticsearch.thread_pool.merge.queue": ("gauge", "thread_pool.merge.queue"),
        "elasticsearch.thread_pool.merge.rejected": ("gauge", "thread_pool.merge.rejected"),
        "elasticsearch.thread_pool.percolate.active": ("gauge", "thread_pool.percolate.active"),
        "elasticsearch.thread_pool.percolate.threads": ("gauge", "thread_pool.percolate.threads"),
        "elasticsearch.thread_pool.percolate.queue": ("gauge", "thread_pool.percolate.queue"),
        "elasticsearch.thread_pool.percolate.rejected": ("gauge", "thread_pool.percolate.rejected"),
        "elasticsearch.thread_pool.refresh.active": ("gauge", "thread_pool.refresh.active"),
        "elasticsearch.thread_pool.refresh.threads": ("gauge", "thread_pool.refresh.threads"),
        "elasticsearch.thread_pool.refresh.queue": ("gauge", "thread_pool.refresh.queue"),
        "elasticsearch.thread_pool.refresh.rejected": ("gauge", "thread_pool.refresh.rejected"),
        "elasticsearch.thread_pool.search.active": ("gauge", "thread_pool.search.active"),
        "elasticsearch.thread_pool.search.threads": ("gauge", "thread_pool.search.threads"),
        "elasticsearch.thread_pool.search.queue": ("gauge", "thread_pool.search.queue"),
        "elasticsearch.thread_pool.search.rejected": ("gauge", "thread_pool.search.rejected"),
        "elasticsearch.thread_pool.snapshot.active": ("gauge", "thread_pool.snapshot.active"),
        "elasticsearch.thread_pool.snapshot.threads": ("gauge", "thread_pool.snapshot.threads"),
        "elasticsearch.thread_pool.snapshot.queue": ("gauge", "thread_pool.snapshot.queue"),
        "elasticsearch.thread_pool.snapshot.rejected": ("gauge", "thread_pool.snapshot.rejected"),
        "elasticsearch.http.current_open": ("gauge", "http.current_open"),
        "elasticsearch.http.total_opened": ("gauge", "http.total_opened"),
        "jvm.gc.concurrent_mark_sweep.count": ("gauge", "jvm.gc.collectors.ConcurrentMarkSweep.collection_count"),
        "jvm.gc.concurrent_mark_sweep.collection_time": ("gauge", "jvm.gc.collectors.ConcurrentMarkSweep.collection_time_in_millis", lambda v: float(v) / 1000),
        "jvm.gc.par_new.count": ("gauge", "jvm.gc.collectors.ParNew.collection_count"),
        "jvm.gc.par_new.collection_time": ("gauge", "jvm.gc.collectors.ParNew.collection_time_in_millis", lambda v: float(v) / 1000),
        "jvm.mem.heap_committed": ("gauge", "jvm.mem.heap_committed_in_bytes"),
        "jvm.mem.heap_used": ("gauge", "jvm.mem.heap_used_in_bytes"),
        "jvm.mem.non_heap_committed": ("gauge", "jvm.mem.non_heap_committed_in_bytes"),
        "jvm.mem.non_heap_used": ("gauge", "jvm.mem.non_heap_used_in_bytes"),
        "jvm.threads.count": ("gauge", "jvm.threads.count"),
        "jvm.threads.peak_count": ("gauge", "jvm.threads.peak_count"),
        "elasticsearch.number_of_nodes": ("gauge", "number_of_nodes"),
        "elasticsearch.number_of_data_nodes": ("gauge", "number_of_data_nodes"),
        "elasticsearch.active_primary_shards": ("gauge", "active_primary_shards"),
        "elasticsearch.active_shards": ("gauge", "active_shards"),
        "elasticsearch.relocating_shards": ("gauge", "relocating_shards"),
        "elasticsearch.initializing_shards": ("gauge", "initializing_shards"),
        "elasticsearch.unassigned_shards": ("gauge", "unassigned_shards"),
        "elasticsearch.cluster_status": ("gauge", "status", lambda v: {"red": 0, "yellow": 1, "green": 2}.get(v, -1)),
    }

    def __init__(self, name, init_config, agent_config):
        AgentCheck.__init__(self, name, init_config, agent_config)

        # Host status needs to persist across all checks
        self.cluster_status = {}

    def check(self, instance):
        config_url = instance.get('url')
        if config_url is None:
            raise Exception("An url must be specified")

        # Load basic authentication configuration, if available.
        username, password = instance.get('username'), instance.get('password')
        if username and password:
            auth = (username, password)
        else:
            auth = None

        # Support URLs that have a path in them from the config, for
        # backwards-compatibility.
        parsed = urlparse.urlparse(config_url)
        if parsed[2] != "":
            config_url = "%s://%s" % (parsed[0], parsed[1])

        # Tag by URL so we can differentiate the metrics from multiple instances
        dimensions = self._set_dimensions({'url': config_url}, instance)

        # Check ES version for this instance and define parameters (URLs and metrics) accordingly
        version = self._get_es_version(config_url, auth)
        self._define_params(version)

        # Load stats data.
        url = urlparse.urljoin(config_url, self.STATS_URL)
        stats_data = self._get_data(url, auth)
        self._process_stats_data(config_url, stats_data, auth, dimensions=dimensions)

        # Load the health data.
        url = urlparse.urljoin(config_url, self.HEALTH_URL)
        health_data = self._get_data(url, auth)
        self._process_health_data(config_url, health_data, dimensions=dimensions)

    def _get_es_version(self, config_url, auth=None):
        """Get the running version of Elastic Search.

        """

        try:
            data = self._get_data(config_url, auth)
            version = map(int, data['version']['number'].split('.'))
        except Exception as e:
            self.log.warn("Error while trying to get Elasticsearch version from %s %s" %
                          (config_url, str(e)))
            version = [0, 0, 0]

        self.log.debug("Elasticsearch version is %s" % version)
        return version

    def _define_params(self, version):
        """Define the set of URLs and METRICS to use depending on the running ES version.

        """

        if version >= [0, 90, 10]:
            # ES versions 0.90.10 and above
            # Metrics architecture changed starting with version 0.90.10
            self.HEALTH_URL = "/_cluster/health?pretty=true"
            self.STATS_URL = "/_nodes/stats?all=true"
            self.NODES_URL = "/_nodes?network=true"

            additional_metrics = {
                "elasticsearch.search.fetch.open_contexts": ("gauge", "indices.search.open_contexts"),
                "elasticsearch.cache.filter.evictions": ("gauge", "indices.filter_cache.evictions"),
                "elasticsearch.cache.filter.size": ("gauge", "indices.filter_cache.memory_size_in_bytes"),
                "elasticsearch.id_cache.size": ("gauge", "indices.id_cache.memory_size_in_bytes"),
                "elasticsearch.fielddata.size": ("gauge", "indices.fielddata.memory_size_in_bytes"),
                "elasticsearch.fielddata.evictions": ("gauge", "indices.fielddata.evictions")
            }

        else:
            # ES version 0.90.9 and below
            self.HEALTH_URL = "/_cluster/health?pretty=true"
            self.STATS_URL = "/_cluster/nodes/stats?all=true"
            self.NODES_URL = "/_cluster/nodes?network=true"

            additional_metrics = {
                "elasticsearch.cache.field.evictions": ("gauge", "indices.cache.field_evictions"),
                "elasticsearch.cache.field.size": ("gauge", "indices.cache.field_size_in_bytes"),
                "elasticsearch.cache.filter.count": ("gauge", "indices.cache.filter_count"),
                "elasticsearch.cache.filter.evictions": ("gauge", "indices.cache.filter_evictions"),
                "elasticsearch.cache.filter.size": ("gauge", "indices.cache.filter_size_in_bytes"),
                "elasticsearch.thread_pool.cache.active": ("gauge", "thread_pool.cache.active"),
                "elasticsearch.thread_pool.cache.threads": ("gauge", "thread_pool.cache.threads"),
                "elasticsearch.thread_pool.cache.queue": ("gauge", "thread_pool.cache.queue"),
                "elasticsearch.thread_pool.cache.rejected": ("gauge", "thread_pool.cache.rejected"),
                "jvm.gc.collection_count": ("gauge", "jvm.gc.collection_count"),
                "jvm.gc.collection_time": ("gauge", "jvm.gc.collection_time_in_millis", lambda v: float(v) / 1000),
                "jvm.gc.copy.count": ("gauge", "jvm.gc.collectors.Copy.collection_count"),
                "jvm.gc.copy.collection_time": ("gauge", "jvm.gc.collectors.Copy.collection_time_in_millis", lambda v: float(v) / 1000)
            }

        self.METRICS.update(additional_metrics)

    def _get_data(self, url, auth=None):
        """Hit a given URL and return the parsed json

        `auth` is a tuple of (username, password) or None
        """
        req = urllib2.Request(url, None, headers(self.agent_config))
        if auth:
            add_basic_auth(req, *auth)
        request = urllib2.urlopen(req)
        response = request.read()
        return json.loads(response)

    def _process_stats_data(self, config_url, data, auth, dimensions=None):
        for node in data['nodes']:
            node_data = data['nodes'][node]

            def process_metric(metric, xtype, path, xform=None):
                # closure over node_data
                self._process_metric(node_data, metric, path, xform, dimensions=dimensions)

            # On newer version of ES it's "host" not "hostname"
            node_hostname = node_data.get('hostname', node_data.get('host', None))

            if node_hostname is not None:
                # For ES >= 0.19
                hostnames = (
                    self.hostname.decode('utf-8'),
                    socket.gethostname().decode('utf-8'),
                    socket.getfqdn().decode('utf-8'),
                    socket.gethostbyname(socket.gethostname()).decode('utf-8')
                )
                if node_hostname.decode('utf-8') in hostnames:
                    for metric in self.METRICS:
                        # metric description
                        desc = self.METRICS[metric]
                        process_metric(metric, *desc)
            else:
                # ES < 0.19
                # Fetch interface address from ifconfig or ip addr and check
                # against the primary IP from ES
                try:
                    nodes_url = urlparse.urljoin(config_url, self.NODES_URL)
                    primary_addr = self._get_primary_addr(nodes_url, node, auth)
                except NodeNotFound:
                    # Skip any nodes that aren't found
                    continue
                if self._host_matches_node(primary_addr):
                    for metric in self.METRICS:
                        # metric description
                        desc = self.METRICS[metric]
                        process_metric(metric, *desc)

    def _get_primary_addr(self, url, node_name, auth):
        """Returns a list of primary interface addresses as seen by ES.

        Used in ES < 0.19
        """
        req = urllib2.Request(url, None, headers(self.agent_config))
        # Load basic authentication configuration, if available.
        if auth:
            add_basic_auth(req, *auth)
        request = urllib2.urlopen(req)
        response = request.read()
        data = json.loads(response)

        if node_name in data['nodes']:
            node = data['nodes'][node_name]
            if 'network' in node\
                    and 'primary_interface' in node['network']\
                    and 'address' in node['network']['primary_interface']:
                return node['network']['primary_interface']['address']

        raise NodeNotFound()

    @staticmethod
    def _host_matches_node(primary_addrs):
        """For < 0.19, check if the current host matches the IP given in the

        cluster nodes check `/_cluster/nodes`. Uses `ip addr` on Linux and
        `ifconfig` on Mac
        """
        if sys.platform == 'darwin':
            ifaces = subprocess.Popen(['ifconfig'], stdout=subprocess.PIPE)
        else:
            ifaces = subprocess.Popen(['ip', 'addr'], stdout=subprocess.PIPE)
        grepper = subprocess.Popen(['grep', 'inet'], stdin=ifaces.stdout,
                                   stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        ifaces.stdout.close()
        out, err = grepper.communicate()

        # Capture the list of interface IPs
        ips = []
        for iface in out.split("\n"):
            iface = iface.strip()
            if iface:
                ips.append(iface.split(' ')[1].split('/')[0])

        # Check the interface addresses against the primary address
        return primary_addrs in ips

    def _process_metric(self, data, metric, path, xform=None, dimensions=None):
        """data: dictionary containing all the stats

        metric: datadog metric
        path: corresponding path in data, flattened, e.g. thread_pool.bulk.queue
        xfom: a lambda to apply to the numerical value
        """
        value = data

        # Traverse the nested dictionaries
        for key in path.split('.'):
            if value is not None:
                value = value.get(key, None)
            else:
                break

        if value is not None:
            if xform:
                value = xform(value)
            if self.METRICS[metric][0] == "gauge":
                self.gauge(metric, value, dimensions=dimensions)
            else:
                self.rate(metric, value, dimensions=dimensions)
        else:
            self._metric_not_found(metric, path)

    def _process_health_data(self, config_url, data, dimensions=None):
        if self.cluster_status.get(config_url, None) is None:
            self.cluster_status[config_url] = data['status']

        if data['status'] != self.cluster_status.get(config_url):
            self.cluster_status[config_url] = data['status']

        def process_metric(metric, xtype, path, xform=None):
            # closure over data
            self._process_metric(data, metric, path, xform, dimensions=dimensions)

        for metric in self.METRICS:
            # metric description
            desc = self.METRICS[metric]
            process_metric(metric, *desc)

    def _metric_not_found(self, metric, path):
        self.log.debug("Metric not found: %s -> %s", path, metric)



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP

import os
import shutil
import tempfile
import unittest

from monasca_setup.detection.plugins import json_plugin


class TestJsonPlugin(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.plugin_obj = json_plugin.JsonPlugin('temp_dir')
        self.varcachedir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.varcachedir)

    def test_var_cache_exists(self):
        json_plugin.VAR_CACHE_DIR = self.varcachedir
        self.plugin_obj._detect()
        result = self.plugin_obj.build_config()
        self.assertTrue(self.plugin_obj.available)
        self.assertEqual(result['json_plugin']['instances'],
                         [{'name': self.varcachedir,
                            'metrics_dir': self.varcachedir}])

    def test_var_cache_not_exists(self):
        json_plugin.VAR_CACHE_DIR = os.path.join(self.varcachedir, 'dummy')
        self.plugin_obj._detect()
        self.assertFalse(self.plugin_obj.available)

    def test_dependencies_installed(self):
        self.assertTrue(self.plugin_obj.dependencies_installed())



# (C) Copyright 2016 Hewlett Packard Enterprise Development LP

import monasca_setup.detection


class FreezerAPI(monasca_setup.detection.ServicePlugin):
    """Detect a running Freezer API server."""

    def __init__(self, template_dir, overwrite=True, args=None):
        service_params = {
            'args': args,
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': 'backup',
            'process_names': ['freezer-api'],
            'service_api_url': 'http://127.0.0.1:9090/healthcheck',
            'search_pattern': '.*OK.*'
        }

        super(FreezerAPI, self).__init__(service_params)


class FreezerScheduler(monasca_setup.detection.ServicePlugin):
    """Detect a running Freezer client-side scheduler."""

    def __init__(self, template_dir, overwrite=True, args=None):
        service_params = {
            'args': args,
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': 'backup',
            'process_names': ['freezer-scheduler'],
            'service_api_url': None,
            'search_pattern': None
        }

        super(FreezerScheduler, self).__init__(service_params)



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development Company LP

import logging
import os
import urllib2

import monasca_setup.agent_config
import monasca_setup.detection

log = logging.getLogger(__name__)

# Process name is apache2 on Debian derivatives, on RHEL derivatives it's httpd
APACHE_PROCESS_NAMES = ('apache2', 'httpd')
DEFAULT_APACHE_CONFIG = '/root/.apache.cnf'
CONFIG_ARG_KEYS = set(["url", "user", "password", "use_server_status_metrics"])


class Apache(monasca_setup.detection.Plugin):

    """Detect Apache web server daemons and setup configuration to monitor apache.

        This plugin will by default setup process check metrics for the apache process,
        and setup server-status metrics using the input url.  If only process check
        metrics are desired, the use_server_status_metrics argument can be passed in
        with a value of false.
        This plugin needs user/password for apache if using the server-status metrics
        when security is setup on the web server.
        This plugin accepts arguments and if none are provided it will attempt to read
        the default configuration file (/root/.apache.cnf) in a format such as:
        [client]
            url=http://localhost/server-status?auto
            user=guest
            password=guest
    """

    def __init__(self, *args, **kwargs):
        self._apache_process_name = 'apache2'
        super(Apache, self).__init__(*args, **kwargs)

    def _detect(self):
        """Run detection, set self.available True if the service is detected.

        """
        process_exists = False
        for proc_name in APACHE_PROCESS_NAMES:
            if monasca_setup.detection.find_process_name(proc_name) is not None:
                self._apache_process_name = proc_name
                process_exists = True
        has_args_or_config_file = (self.args is not None or
                                   os.path.isfile(DEFAULT_APACHE_CONFIG))
        self.available = process_exists and has_args_or_config_file
        if not self.available:
            if not process_exists:
                log.error('Apache process does not exist.')
            elif not has_args_or_config_file:
                log.error(('Apache process exists but '
                           'configuration file was not found and '
                           'no arguments were given.'))

    def _read_apache_config(self, config_location):
        # Read the apache config file to extract the needed variables.
        client_section = False
        apache_url = None
        apache_user = None
        apache_pass = None
        use_server_status_metrics = True
        try:
            with open(config_location, "r") as config_file:
                for row in config_file:
                    if "[client]" in row:
                        client_section = True
                        continue
                    if client_section:
                        if "url=" in row:
                            apache_url = row.split("=")[1].strip()
                        if "user=" in row:
                            apache_user = row.split("=")[1].strip()
                        if "password=" in row:
                            apache_pass = row.split("=")[1].strip()
                        if "use_server_status_metrics" in row:
                            use_server_status_metrics = row.split("=")[1].strip()
        except IOError:
            log.warn("\tUnable to read {:s}".format(config_location))

        return apache_url, apache_user, apache_pass, use_server_status_metrics

    def build_config(self):
        """Build the config as a Plugins object and return.

        """
        config = monasca_setup.agent_config.Plugins()
        # First watch the process
        config.merge(monasca_setup.detection.watch_process(
            [self._apache_process_name], 'apache'))
        log.info("\tWatching the apache webserver process.")

        error_msg = '\n\t*** The Apache plugin is not configured ***\n\tPlease correct and re-run monasca-setup.'
        # Attempt login, requires either an empty root password from localhost
        # or relying on a configured /root/.apache.cnf
        if self.dependencies_installed():
            # If any of the exact keys are present, use them.
            if self.args and self.args.viewkeys() & CONFIG_ARG_KEYS:
                log.info("Attempting to use command args")
                apache_url = self.args.get("url", None)
                apache_user = self.args.get("user", None)
                apache_pass = self.args.get("password", None)
                use_server_status_metrics = self.args.get('use_server_status_metrics', True)
            elif self.args and self.args.get("apache_config_file"):
                config_file = self.args.get("apache_config_file")
                apache_url, apache_user, apache_pass, use_server_status_metrics = \
                    self._read_apache_config(config_file)
            else:
                apache_url, apache_user, apache_pass, use_server_status_metrics = \
                    self._read_apache_config(DEFAULT_APACHE_CONFIG)
            if type(use_server_status_metrics) is str:
                use_server_status_metrics = (use_server_status_metrics.lower() == 'true')

            if use_server_status_metrics:
                if not apache_url:
                    missing_url_msg = ('\tNo server-status url specified.' + error_msg)
                    log.error(missing_url_msg)
                    raise Exception(missing_url_msg)
            else:
                log.info("\tWatching the apache process only.")
                return config

            if apache_user and apache_pass:
                password_mgr = urllib2.HTTPPasswordMgrWithDefaultRealm()
                password_mgr.add_password(None,
                                          apache_url,
                                          apache_user,
                                          apache_pass)
                handler = urllib2.HTTPBasicAuthHandler(password_mgr)
            else:
                if 'https' in apache_url:
                    handler = urllib2.HTTPSHandler()
                else:
                    handler = urllib2.HTTPHandler()

            opener = urllib2.build_opener(handler)

            try:
                request = opener.open(apache_url)
                response = request.read()
                request.close()
                if 'Total Accesses:' in response:
                    instance_vars = {'name': apache_url, 'apache_status_url': apache_url}
                    if apache_user and apache_pass:
                        instance_vars.update({'apache_user': apache_user,
                                              'apache_password': apache_pass})
                    config['apache'] = {'init_config': None, 'instances': [instance_vars]}
                    log.info("\tSuccessfully setup Apache plugin.")
                else:
                    log.warn('Unable to access the Apache server-status URL;' + error_msg)
            except urllib2.URLError as e:
                exception_msg = (
                    '\tError {0} received when accessing url {1}.'.format(e.reason, apache_url) +
                    '\n\tPlease ensure the Apache web server is running and your configuration ' +
                    'information is correct.' + error_msg)
                log.error(exception_msg)
                raise Exception(exception_msg)
            except Exception as e:
                exception_msg = (
                    'Error received when accessing url {0} exception {1}'.format(apache_url, e) +
                    error_msg)
                log.error(exception_msg)
                raise Exception(exception_msg)
        else:
            log.error('\tThe dependencies for Apache Web Server are not installed or unavailable.' + error_msg)

        return config

    def dependencies_installed(self):
        # No real dependencies to check
        return True



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP

"""Classes to aid in configuration of the agent."""

import collections
import logging
import os
import pwd
import yaml

log = logging.getLogger(__name__)


class Plugins(collections.defaultdict):

    """A container for the plugin configurations used by the monasca-agent.

        This is essentially a defaultdict(dict) but put into a class primarily to make the interface clear, also
        to add a couple of helper methods.
        Each plugin config is stored with the key being its config name (excluding .yaml).
        The value a dict which will convert to yaml.
    """

    def __init__(self):
        super(Plugins, self).__init__(dict)

    # todo Possibly enforce the key being a string without .yaml in it.

    def diff(self, other_plugins):
        raise NotImplementedError

    def merge(self, other):
        """Do a deep merge with precedence going to other (as is the case with update).

        """
        # Implemented as a function so it can be used for arbitrary dictionaries not just self, this is needed
        # for the recursive nature of the merge.
        deep_merge(self, other)


def deep_merge(adict, other):
    """A recursive merge of two dictionaries including combining of any lists within the data structure.

    """
    for key, value in other.iteritems():
        if key in adict:
            if isinstance(adict[key], dict) and isinstance(value, dict):
                deep_merge(adict[key], value)
            elif isinstance(adict[key], list) and isinstance(value, list):
                adict[key] += value
        else:
            adict[key] = value


def merge_by_name(first, second):
    """Merge a list of dictionaries replacing any dictionaries with the same 'name' value rather than merging.
        The precedence goes to first.
    """
    first_names = [i['name'] for i in first if 'name' in i]
    for item in second:
        if 'name' not in item or item['name'] not in first_names:
            first.append(item)


def read_plugin_config_from_disk(config_dir, plugin_name):
    """Reads from the Agent on disk configuration the config for a specific plugin
    :param config_dir: Monasca Agent configuration directory
    :param plugin_name: The name of the check plugin
    :return: Dictionary of parsed yaml content
    """
    config_path = os.path.join(config_dir, 'conf.d', plugin_name + '.yaml')
    config = None
    if os.path.exists(config_path):
        with open(config_path, 'r') as config_file:
            config = yaml.safe_load(config_file.read())
    return config


def save_plugin_config(config_dir, plugin_name, user, conf):
    """Writes configuration for plugin_name to disk in the config_dir
    :param config_dir: Monasca Agent configuration directory
    :param plugin_name: The name of the check plugin
    :param user: The username Monasca-agent will run as
    :param conf: The value of the configuration to write to disk
    :return: None
    """
    config_path = os.path.join(config_dir, 'conf.d', plugin_name + '.yaml')

    with open(config_path, 'w') as config_file:
        # The gid is created on service activation which we assume has happened
        config_file.write(yaml.safe_dump(conf,
                                         encoding='utf-8',
                                         allow_unicode=True,
                                         default_flow_style=False))
    stat = pwd.getpwnam(user)

    gid = stat.pw_gid
    uid = stat.pw_uid

    os.chmod(config_path, 0o640)
    os.chown(config_path, uid, gid)


def check_endpoint_changes(value, config):
    """Change urls in config with same path but different protocols into new
       endpoints.
    """
    new_url = value['instances'][0]['url']
    old_urls = [i['url'] for i in config['instances'] if 'url' in i]
    new_path = new_url.split("://")[1]
    old_paths = [url.split("://")[1] for url in old_urls]
    for i, old_path in enumerate(old_paths):
        if old_path == new_path:
            if config['instances'][i]['url'] == config['instances'][i]['name']:
                config['instances'][i]['name'] = new_url
            config['instances'][i]['url'] = new_url
    return config


def delete_from_config(args, config, file_path, plugin_name):
    if args.dry_run:
        info_msg = ("Changes would be made to the config file {0}".format(file_path))
    else:
        if len(config['instances']) == 0:
            info_msg = ("Removing configuration file {0} it is no longer needed.".format(file_path))
            os.remove(file_path)
        else:
            info_msg = ("Saving changes to configuration file {0}.".format(file_path))
            save_plugin_config(args.config_dir, plugin_name, args.user, config)
    log.info(info_msg)



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP

from hashlib import md5
import json
import urllib2


def post_headers(payload):
    return {
        'User-Agent': 'Mon/Agent',
        'Content-Type': 'application/json',
        'Accept': 'text/html, */*',
        'Content-MD5': md5(payload).hexdigest()
    }


def http_emitter(message, log, url):
    """Send payload
    """

    log.debug('http_emitter: attempting postback to ' + url)

    # Post back the data
    partial_payload = []
    for measurement in message:
        partial_payload.append(measurement)

    payload = json.dumps(partial_payload)
    url = "%s/intake" % url
    headers = post_headers(payload)

    try:
        # Make sure no proxy is autodetected for this localhost connection
        proxy_handler = urllib2.ProxyHandler({})
        # Should this be installed as the default opener and reused?
        opener = urllib2.build_opener(proxy_handler)
        request = urllib2.Request(url, payload, headers)
        response = None
        try:
            response = opener.open(request)
            log.debug('http_emitter: postback response: ' + str(response.read()))
        except Exception as exc:
            log.error("""Forwarder at {0} is down or not responding...
                      Error is {1}
                      Please restart the monasca-agent.""".format(url, repr(exc)))
        finally:
            if response:
                response.close()
    except urllib2.HTTPError as e:
        if e.code == 202:
            log.debug("http payload accepted")
        else:
            raise



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import monasca_setup.detection


class Octavia(monasca_setup.detection.ServicePlugin):

    """Detect Octavia  daemons and setup configuration to monitor them.

    """

    def __init__(self, template_dir, overwrite=True, args=None):
        service_params = {
            'args': args,
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': 'octavia',
            'process_names': ['octavia-api', 'octavia-worker',
                              'octavia-health-manager', 'octavia-housekeeping'],
            'service_api_url': 'http://localhost:9876',
            'search_pattern': '.*200 OK.*',
        }

        super(Octavia, self).__init__(service_params)



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development Company LP

from args_plugin import ArgsPlugin
from plugin import Plugin
from service_plugin import ServicePlugin
from utils import find_process_cmdline
from utils import find_process_name
from utils import find_process_service
from utils import watch_process
from utils import watch_process_by_username



# (c) Copyright 2015-2016 Hewlett Packard Enterprise Development LP

import ConfigParser
import grp
import logging
import os
import pwd
from shutil import copy
import subprocess
import sys

from monasca_agent.common.psutil_wrapper import psutil
import monasca_setup.agent_config
from monasca_setup.detection import Plugin


log = logging.getLogger(__name__)

# Directory to use for instance and metric caches (preferred tmpfs "/dev/shm")
cache_dir = "/dev/shm"
# Maximum age of instance cache before automatic refresh (in seconds)
nova_refresh = 60 * 60 * 4  # Four hours
# Probation period before metrics are gathered for a VM (in seconds)
vm_probation = 60 * 5  # Five minutes
# List of instance metadata keys to be sent as dimensions
# By default 'scale_group' metadata is used here for supporting auto
# scaling in Heat.
metadata = ['scale_group']
# Include scale group dimension for customer metrics.
customer_metadata = ['scale_group']
# List 'ping' commands (paths and parameters) in order of preference.
# The plugin will use the first fuctional command. 127.0.0.1 will be appended.
ping_options = [["/usr/bin/fping", "-n", "-c1", "-t250", "-q"],
                ["/sbin/fping", "-n", "-c1", "-t250", "-q"],
                ["/bin/ping", "-n", "-c1", "-w1", "-q"]]
# Path to 'ip' command (needed to execute ping within network namespaces)
ip_cmd = "/sbin/ip"
# How many ping commands to run concurrently
default_max_ping_concurrency = 8
# Disk metrics can be collected at a larger interval than other vm metrics
default_disk_collection_period = 0
# VNIC metrics can be collected at a larger interval than other vm metrics
default_vnic_collection_period = 0

# Arguments which should be written as integers, not strings
INT_ARGS = ['disk_collection_period', 'vnic_collection_period',
            'max_ping_concurrency', 'nova_refresh', 'vm_probation']


class Libvirt(Plugin):
    """Configures VM monitoring through Nova"""

    def _detect(self):
        """Set self.available True if the process and config file are detected
        """
        # Detect Agent's OS username by getting the group owner of confg file
        try:
            gid = os.stat('/etc/monasca/agent/agent.yaml').st_gid
            self.agent_user = grp.getgrgid(gid)[0]
        except OSError:
            self.agent_user = None
        # Try to detect the location of the Nova configuration file.
        # Walk through the list of processes, searching for 'nova-compute'
        # process with 'nova.conf' inside one of the parameters
        nova_conf = None
        for proc in psutil.process_iter():
            try:
                cmd = proc.as_dict(['cmdline'])['cmdline']
                if len(cmd) > 2 and 'python' in cmd[0] and 'nova-compute' in cmd[1]:
                    conf_indexes = [cmd.index(y) for y in cmd if 'nova.conf' in y]
                    if not conf_indexes:
                        continue
                    param = conf_indexes[0]
                    if '=' in cmd[param]:
                        nova_conf = cmd[param].split('=')[1]
                    else:
                        nova_conf = cmd[param]
            except (IOError, psutil.NoSuchProcess):
                # Process has already terminated, ignore
                continue
        if (nova_conf is not None and os.path.isfile(nova_conf)):
            self.available = True
            self.nova_conf = nova_conf

    def build_config(self):
        """Build the config as a Plugins object and return back.
        """
        config = monasca_setup.agent_config.Plugins()

        if self.dependencies_installed():
            nova_cfg = ConfigParser.SafeConfigParser()
            log.info("\tUsing nova configuration file {0}".format(self.nova_conf))
            nova_cfg.read(self.nova_conf)
            # Which configuration options are needed for the plugin YAML?
            # Use a dict so that they can be renamed later if necessary
            cfg_needed = {'admin_user': 'admin_user',
                          'admin_password': 'admin_password',
                          'admin_tenant_name': 'admin_tenant_name'}
            cfg_section = 'keystone_authtoken'

            # Handle Devstack's slightly different nova.conf names
            if (nova_cfg.has_option(cfg_section, 'username')
               and nova_cfg.has_option(cfg_section, 'password')
               and nova_cfg.has_option(cfg_section, 'project_name')):
                cfg_needed = {'admin_user': 'username',
                              'admin_password': 'password',
                              'admin_tenant_name': 'project_name'}

            # Start with plugin-specific configuration parameters
            init_config = {'cache_dir': cache_dir,
                           'nova_refresh': nova_refresh,
                           'vm_probation': vm_probation,
                           'metadata': metadata,
                           'customer_metadata': customer_metadata,
                           'max_ping_concurrency': default_max_ping_concurrency,
                           'disk_collection_period': default_disk_collection_period,
                           'vnic_collection_period': default_vnic_collection_period}

            # Set default parameters for included checks
            init_config['vm_cpu_check_enable'] = self.literal_eval('True')
            init_config['vm_disks_check_enable'] = self.literal_eval('True')
            init_config['vm_network_check_enable'] = self.literal_eval('True')
            init_config['vm_ping_check_enable'] = self.literal_eval('True')
            init_config['vm_extended_disks_check_enable'] = self.literal_eval('False')

            for option in cfg_needed:
                init_config[option] = nova_cfg.get(cfg_section, cfg_needed[option])

            # Create an identity URI (again, slightly different for Devstack)
            if nova_cfg.has_option(cfg_section, 'auth_url'):
                init_config['identity_uri'] = "{0}/v2.0".format(nova_cfg.get(cfg_section, 'auth_url'))
            else:
                init_config['identity_uri'] = "{0}/v2.0".format(nova_cfg.get(cfg_section, 'identity_uri'))

            # Verify requirements to enable ping checks
            init_config['ping_check'] = self.literal_eval('False')
            if self.agent_user is None:
                log.warn("\tUnable to determine agent user.  Skipping ping checks.")
            else:
                try:
                    from neutronclient.v2_0 import client

                    # Copy system 'ip' command to local directory
                    copy(ip_cmd, sys.path[0])
                    # Restrict permissions on the local 'ip' command
                    os.chown("{0}/ip".format(sys.path[0]), pwd.getpwnam(self.agent_user).pw_uid, 0)
                    os.chmod("{0}/ip".format(sys.path[0]), 0o700)
                    # Set capabilities on 'ip' which will allow
                    # self.agent_user to exec commands in namespaces
                    setcap_cmd = ['/sbin/setcap', 'cap_sys_admin+ep',
                                  "{0}/ip".format(sys.path[0])]
                    subprocess.Popen(setcap_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    # Verify that the capabilities were set
                    setcap_cmd.extend(['-v', '-q'])
                    subprocess.check_call(setcap_cmd)
                    # Look for the best ping command
                    for ping_cmd in ping_options:
                        if os.path.isfile(ping_cmd[0]):
                            init_config['ping_check'] = "{0}/ip netns exec NAMESPACE {1}".format(sys.path[0],
                                                                                                 ' '.join(ping_cmd))
                            log.info("\tEnabling ping checks using {0}".format(ping_cmd[0]))
                            break
                    if init_config['ping_check'] is False:
                        log.warn("\tUnable to find suitable ping command, disabling ping checks.")
                except ImportError:
                    log.warn("\tneutronclient module missing, required for ping checks.")
                    pass
                except IOError:
                    log.warn("\tUnable to copy {0}, ping checks disabled.".format(ip_cmd))
                    pass
                except (subprocess.CalledProcessError, OSError):
                    log.warn("\tUnable to set up ping checks, setcap failed ({0})".format(' '.join(setcap_cmd)))
                    pass

            # Handle monasca-setup detection arguments, which take precedence
            if self.args:
                for arg in self.args:
                    if arg in INT_ARGS:
                        value = self.args[arg]
                        try:
                            init_config[arg] = int(value)
                        except ValueError:
                            log.warn("\tInvalid integer value '{0}' for parameter {1}, ignoring value"
                                     .format(value, arg))
                    else:
                        init_config[arg] = self.literal_eval(self.args[arg])

            config['libvirt'] = {'init_config': init_config,
                                 'instances': []}

        return config

    def dependencies_installed(self):
        try:
            import json
            import monasca_agent.collector.virt.inspector
            import time

            from netaddr import all_matching_cidrs
            from novaclient import client
        except ImportError:
            log.warn("\tDependencies not satisfied; plugin not configured.")
            return False
        if os.path.isdir(cache_dir) is False:
            log.warn("\tCache directory {} not found;" +
                     " plugin not configured.".format(cache_dir))
            return False
        return True



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP

import ConfigParser
import logging
import os
import psutil
import re

import monasca_setup.agent_config
import monasca_setup.detection

log = logging.getLogger(__name__)

# Maximum time to wait before updating neutron router cache(in seconds)
neutron_refresh = 60 * 60 * 4  # Four hours
# Directory to use for metric caches
cache_dir = "/dev/shm"
# ovs-vsctl command needs sudo privileges to connect to
# /var/run/openvswitch/db.sock which is created by ovsdb-server.
ovs_cmd = "sudo /usr/bin/ovs-vsctl"
# If set, will submit network metrics in bits
network_use_bits = False
# Regular expression for interface types
included_interface_re = 'qg.*|vhu.*|sg.*'
# If set, will submit raw counters from ovs-vsctl command output for the given
# network interface
use_absolute_metrics = True
# If set, will submit the rate metrics
use_rate_metrics = True
# If set, will submit the health metrics
use_health_metrics = True
# If set, router max capacity metrics will be published
publish_router_capacity = False
# Acceptable arguments
acceptable_args = ['admin_user', 'admin_password', 'admin_tenant_name',
                   'identity_uri', 'cache_dir', 'neutron_refresh', 'ovs_cmd',
                   'network_use_bits', 'check_router_ha', 'region_name',
                   'included_interface_re', 'conf_file_path', 'use_absolute_metrics',
                   'use_rate_metrics', 'use_health_metrics', 'publish_router_capacity']
# Arguments which must be ignored if provided
ignorable_args = ['admin_user', 'admin_password', 'admin_tenant_name',
                  'identity_uri', 'region_name', 'conf_file_path']
# Regular expression to match the URI version
uri_version_re = re.compile('.*v2.0|.*v3.0|.*v1|.*v2')


class Ovs(monasca_setup.detection.Plugin):
    """Detect OVS daemons and setup configuration to monitor

    """
    def _detect(self):
        process_exist = (monasca_setup.detection.
                         find_process_cmdline('neutron-openvsw'))
        has_dependencies = self.dependencies_installed()
        neutron_conf = self.get_ovs_config_file() if process_exist else ''

        self.available = (process_exist is not None and
                          os.path.isfile(neutron_conf) and has_dependencies)
        if not self.available:
            if not process_exist:
                log.error('OVS daemon process does not exist.')
            elif not neutron_conf:
                log.error(('OVS daemon process exists but configuration '
                           'file was not found. Path to file does not exist '
                           'as a process parameter or was not '
                           'passed via args.'))
            elif not has_dependencies:
                log.error(('OVS daemon process exists but required '
                           'dependence python-neutronclient is '
                           'not installed.'))
        else:
            self.neutron_conf = neutron_conf

    def get_ovs_config_file(self):
        neutron_conf = None
        if self.args:
            for arg in self.args:
                if arg == 'conf_file_path':
                    neutron_conf = self.args[arg]
        # Try to detect the location of the Neutron configuration file.
        # Walk through the list of processes, searching for 'neutron'
        # process with 'neutron.conf' inside one of the parameters.
        if not neutron_conf:
            for proc in psutil.process_iter():
                try:
                    proc_dict = proc.as_dict()
                    if proc_dict['name'] == 'neutron-openvsw':
                        cmd = proc_dict['cmdline']
                        neutron_config_params = [param for param in cmd if
                                                 'neutron.conf' in param]
                        if not neutron_config_params:
                            continue
                        if '=' in neutron_config_params[0]:
                            neutron_conf = neutron_config_params[0].split('=')[
                                1]
                        else:
                            neutron_conf = neutron_config_params[0]
                except (IOError, psutil.NoSuchProcess):
                    # Process has already terminated, ignore
                    continue
        return neutron_conf

    def build_config(self):
        """Build the config as a Plugins object and return.

        """
        config = monasca_setup.agent_config.Plugins()
        neutron_cfg = ConfigParser.SafeConfigParser()
        log.info("\tUsing neutron configuration file {0}".format(self.neutron_conf))
        neutron_cfg.read(self.neutron_conf)
        cfg_needed = {'admin_user': 'admin_user',
                      'admin_password': 'admin_password',
                      'admin_tenant_name': 'admin_tenant_name'}
        cfg_section = 'keystone_authtoken'

        # Handle Devstack's slightly different neutron.conf names
        if (
           neutron_cfg.has_option(cfg_section, 'username') and
           neutron_cfg.has_option(cfg_section, 'password') and
           neutron_cfg.has_option(cfg_section, 'project_name')):
            cfg_needed = {'admin_user': 'username',
                          'admin_password': 'password',
                          'admin_tenant_name': 'project_name'}

        # Start with plugin-specific configuration parameters
        init_config = {'cache_dir': cache_dir,
                       'neutron_refresh': neutron_refresh,
                       'ovs_cmd': ovs_cmd,
                       'network_use_bits': network_use_bits,
                       'included_interface_re': included_interface_re,
                       'use_absolute_metrics': use_absolute_metrics,
                       'use_rate_metrics': use_rate_metrics,
                       'use_health_metrics': use_health_metrics,
                       'publish_router_capacity': publish_router_capacity}

        for option in cfg_needed:
            init_config[option] = neutron_cfg.get(cfg_section, cfg_needed[option])

        uri_version = 'v2.0'
        if neutron_cfg.has_option(cfg_section, 'auth_version'):
            uri_version = str(neutron_cfg.get(cfg_section, 'auth_version'))

        # Create an identity URI (again, slightly different for Devstack)
        if neutron_cfg.has_option(cfg_section, 'auth_url'):
            if re.match(uri_version_re, str(neutron_cfg.get(cfg_section, 'auth_url'))):
                uri_version = ''
            init_config['identity_uri'] = "{0}/{1}".format(neutron_cfg.get(cfg_section, 'auth_url'), uri_version)
        else:
            init_config['identity_uri'] = "{0}/{1}".format(neutron_cfg.get(cfg_section, 'identity_uri'), uri_version)

        # Create an region_name (again, slightly different for Devstack)
        if neutron_cfg.has_option('service_auth', 'region'):
            init_config['region_name'] = str(neutron_cfg.get('service_auth', 'region'))
        else:
            try:
                init_config['region_name'] = str(neutron_cfg.get('nova', 'region_name'))
            except ConfigParser.NoOptionError:
                log.debug(('Option region_name was not found in nova '
                           'section, nova_region_name option from '
                           'DEFAULT section will be used.'))
                init_config['region_name'] = str(neutron_cfg.get('DEFAULT',
                                                                 'nova_region_name'))

        # Handle monasca-setup detection arguments, which take precedence
        if self.args:
            for arg in self.args:
                if arg in acceptable_args and arg not in ignorable_args:
                    if arg == 'included_interface_re':
                        try:
                            re.compile(self.args[arg])
                        except re.error as e:
                            exception_msg = (
                                "Invalid regular expression given for "
                                "'included_interface_re'. {0}.".format(e))
                            log.exception(exception_msg)
                            raise Exception(exception_msg)

                    init_config[arg] = self.literal_eval(self.args[arg])
                elif arg in ignorable_args:
                    log.warn("Argument '{0}' is ignored; Fetching {0} from "
                             "neutron configuration file.".format(arg))
                else:
                    log.warn("Invalid argument '{0}' "
                             "has been provided!!!".format(arg))

        config['ovs'] = {'init_config': init_config,
                         'instances': []}

        return config

    def dependencies_installed(self):
        try:
            import neutronclient.v2_0.client
        except ImportError:
            return False
        return True



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP

import base64


def add_basic_auth(request, username, password):
    """A helper to add basic authentication to a urllib2 request.

    We do this across a variety of checks so it's good to have this in one place.
    """
    auth_str = base64.encodestring('%s:%s' % (username, password)).strip()
    request.add_header('Authorization', 'Basic %s' % auth_str)
    return request


def get_keystone_client(config):
    import keystoneclient.v2_0.client as kc
    kwargs = {
        'username': config.get('admin_user'),
        'project_name': config.get('admin_tenant_name'),
        'password': config.get('admin_password'),
        'auth_url': config.get('identity_uri'),
        'endpoint_type': 'internalURL',
        'region_name': config.get('region_name'),
    }

    return kc.Client(**kwargs)


def get_tenant_name(tenants, tenant_id):
    tenant_name = None
    for tenant in tenants:
        if tenant.id == tenant_id:
            tenant_name = tenant.name
            break
    return tenant_name


def get_tenant_list(config, log):
    tenants = []
    try:
        log.debug("Retrieving Keystone tenant list")
        keystone = get_keystone_client(config)
        tenants = keystone.tenants.list()
    except Exception as e:
        msg = "Unable to get tenant list from keystone: {0}"
        log.error(msg.format(e))

    return tenants



# Copyright 2016 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import random
import unittest

import logging
import mock
import psutil

from monasca_setup.detection.plugins import mon

_PYTHON_CMD_API = ('/opt/monasca-api/bin/gunicorn'
                   ' -n monasca-api'
                   ' -k eventlet --worker-connections=2000 '
                   '--backlog=1000 '
                   '--paste /etc/monasca/api-config.ini -w 9')
_JAVA_CMD_API = ('/usr/bin/java '
                 '-Dfile.encoding=UTF-8 -Xmx128m '
                 '-cp /opt/monasca/monasca-api.jar '
                 'monasca.api.MonApiApplication server '
                 '/etc/monasca/api-config.yml')

_PYTHON_CMD_PERSISTER = ('/opt/monasca-persister/bin/python '
                         '/opt/monasca-persister/lib/python2.7/'
                         'site-packages/monasca_persister/persister.py'
                         ' --config-file /etc/monasca/persister.conf')
_JAVA_CMD_PERSISTER = ('/usr/bin/java -Dfile.encoding=UTF-8 -Xmx128m '
                       '-cp /opt/monasca/monasca-persister.jar '
                       'monasca.persister.PersisterApplication server '
                       '/etc/monasca/persister-config.yml')

_JAVA_YML_CFG_BIT_API = '''
hibernate:
  supportEnabled: {hibernate_enabled}
server:
  adminConnectors:
    - type: {admin_type}
      port: {admin_port}
  applicationConnectors:
    - port: {app_port}
'''

_JAVA_YML_CFG_BIT_PERSISTER = '''
alarmHistoryConfiguration:
    numThreads: {ah_threads}
metricConfiguration:
    numThreads: {mc_threads}
databaseConfiguration:
    databaseType: {db_type}
server:
    adminConnectors:
        - type: {admin_type}
          port: {admin_port}
'''

LOG = logging.getLogger(mon.__name__)


class FakeInetConnection(object):
    def __init__(self, api_port=None):
        if api_port is None:
            api_port = 8070  # default one
        self.laddr = [1, api_port]


class FakeProcesses(object):
    name = None
    cmdLine = None
    inetConnections = [
        FakeInetConnection()
    ]

    def as_dict(self):
        return {'name': FakeProcesses.name,
                'cmdline': FakeProcesses.cmdLine,
                'exe': self.exe()}

    def cmdline(self):
        return self.cmdLine

    def exe(self):
        return self.cmdLine[0]

    def connections(self, *args):
        return FakeProcesses.inetConnections


class TestGetImplLang(unittest.TestCase):
    @mock.patch('psutil.Process')
    def test_should_return_python_lang_for_gunicorn_process(self, proc):
        proc.as_dict.return_value = {'exe': '/opt/monasca-api/bin/gunicorn'}
        self.assertEqual('python', mon._get_impl_lang(proc))

    @mock.patch('psutil.Process')
    def test_should_return_python_lang_for_python_process(self, proc):
        proc.as_dict.return_value = {'exe': '/usr/bin/python'}
        self.assertEqual('python', mon._get_impl_lang(proc))

    @mock.patch('psutil.Process')
    def test_should_return_java_lang_for_java_process(self, proc):
        proc.as_dict.return_value = {'exe': '/usr/bin/java'}
        self.assertEqual('java', mon._get_impl_lang(proc))

    @mock.patch('psutil.Process')
    def test_should_throw_error_for_unknown_impl(self, proc):
        proc.as_dict.return_value = {'exe': '/usr/bin/cat'}
        self.assertRaises(Exception, mon._get_impl_lang, proc)


class TestMonPersisterDetectionPlugin(unittest.TestCase):

    def setUp(self):
        super(TestMonPersisterDetectionPlugin, self).setUp()
        FakeProcesses.name = 'monasca-persister'
        with mock.patch.object(mon.MonPersister, '_detect') as mock_detect:
            self._mon_p = mon.MonPersister('foo')
            self.assertTrue(mock_detect.called)

    @mock.patch(
        'monasca_setup.detection.plugins.mon._MonPersisterJavaHelper')
    def test_should_use_java_helper_if_persister_is_java(self,
                                                              impl_helper):
        FakeProcesses.cmdLine = [_JAVA_CMD_PERSISTER]

        self._mon_p._init_impl_helper = iih = mock.Mock(
            return_value=impl_helper)

        self._detect()

        iih.assert_called_once_with('java')
        self.assertTrue(impl_helper.load_configuration.called_once)

    def test_should_detect_java_persister_has_config(self):
        FakeProcesses.cmdLine = [_JAVA_CMD_PERSISTER]

        yml_cfg = _JAVA_YML_CFG_BIT_PERSISTER.format(
            ah_threads=5,
            mc_threads=5,
            db_type="influxdb",
            admin_type="http",
            admin_port=6666
        )

        with mock.patch(
                "__builtin__.open",
                mock.mock_open(read_data=yml_cfg)) as mf:
            self._detect()
            mf.assert_called_once_with('/etc/monasca/persister-config.yml',
                                       'r')

        self.assertTrue(self._mon_p.available)

    @mock.patch('six.moves.configparser.RawConfigParser')
    def test_should_detect_python_persister_has_config(self, _):
        # NOTE(trebskit) this cannot use mocking the read of the file
        # because when either RawConfigParser or mock_open messes up with
        # reading the one file line. Instead returning empty line,
        # StopIteration is raised and RawConfigParser does not ignore that or
        # catch it
        #
        # to sum it up => ;-(((

        FakeProcesses.cmdLine = [_PYTHON_CMD_PERSISTER]
        self._mon_p._init_impl_helper = mock.Mock(return_value=mock.Mock())

        self._detect()
        self.assertTrue(self._mon_p.available)

    def test_build_java_config(self):
        FakeProcesses.cmdLine = [_JAVA_CMD_PERSISTER]

        # note(trebskit) this is always set to 2
        jvm_metrics_count = 2

        alarm_history_threads = random.randint(1, 5)
        metrics_history_threads = random.randint(1, 5)
        admin_port = random.randint(1000, 10000)

        admin_type = 'http'
        if admin_port % 2 != 0:
            admin_type += 's'

        yml_cfg = _JAVA_YML_CFG_BIT_PERSISTER.format(
            ah_threads=alarm_history_threads,
            mc_threads=metrics_history_threads,
            db_type="influxdb",
            admin_type=admin_type,
            admin_port=admin_port
        )

        with mock.patch(
                "__builtin__.open",
                mock.mock_open(read_data=yml_cfg)) as mf:
            self._detect()
            conf = self._build_config()
            mf.assert_called_once_with('/etc/monasca/persister-config.yml',
                                       'r')

        for key in ('http_check', 'http_metrics', 'process'):
            self.assertIn(key, conf)
            bit = conf[key]
            self.assertIsNotNone(bit)
            self.assertNotEqual({}, bit)

        # detailed assertions

        # http_check
        http_check_instance = conf['http_check']['instances'][0]
        self.assertFalse(http_check_instance['include_content'])
        self.assertEqual('monitoring-monasca-persister healthcheck',
                         http_check_instance['name'])
        self.assertEqual(5, http_check_instance['timeout'])
        self.assertEqual('%s://localhost:%d/healthcheck'
                         % (admin_type, admin_port),
                         http_check_instance['url'])

        # http_metrics
        http_metrics_instance = conf['http_metrics']['instances'][0]
        self.assertEqual('monitoring-monasca-persister metrics',
                         http_metrics_instance['name'])
        self.assertEqual('%s://localhost:%d/metrics'
                         % (admin_type, admin_port),
                         http_metrics_instance['url'])
        hmi_whitelist = http_metrics_instance['whitelist']
        self.assertIsNotNone(hmi_whitelist)
        self.assertEqual(len(hmi_whitelist), (
            alarm_history_threads +
            metrics_history_threads +
            jvm_metrics_count))

        jvm_metrics_found = 0
        ah_metrics_found = 0
        mh_metrics_found = 0

        for entry in hmi_whitelist:
            name = entry['name']
            if 'jvm' in name:
                jvm_metrics_found += 1
            elif 'alarm-state' in name:
                ah_metrics_found += 1
            elif 'metrics-added' in name:
                mh_metrics_found += 1
        self.assertEqual(jvm_metrics_count, jvm_metrics_found)
        self.assertEqual(alarm_history_threads, ah_metrics_found)
        self.assertEqual(metrics_history_threads, mh_metrics_found)

        # process
        process_instance = conf['process']['instances'][0]
        self.assertEqual('monasca-persister', process_instance['name'])
        self.assertFalse(process_instance['exact_match'])
        self.assertTrue(process_instance['detailed'])
        self.assertDictEqual({
            'component': 'monasca-persister',
            'service': 'monitoring'
        }, process_instance['dimensions'])

    def test_build_python_config(self):
        FakeProcesses.cmdLine = [_PYTHON_CMD_PERSISTER]
        self._detect()
        conf = self._build_config()

        for key in ('process',):
            self.assertIn(key, conf)
            bit = conf[key]
            self.assertIsNotNone(bit)
            self.assertNotEqual({}, bit)

        # process
        process_instance = conf['process']['instances'][0]
        self.assertEqual('monasca-persister', process_instance['name'])
        self.assertFalse(process_instance['exact_match'])
        self.assertTrue(process_instance['detailed'])
        self.assertDictEqual({
            'component': 'monasca-persister',
            'service': 'monitoring'
        }, process_instance['dimensions'])

    def _detect(self):
        self._mon_p.available = False
        process_iter = mock.patch.object(psutil, 'process_iter',
                                         return_value=[FakeProcesses()])
        with process_iter as mock_process_iter:
            self._mon_p._detect()
            self.assertTrue(mock_process_iter.called)

    def _build_config(self):
        conf = self._mon_p.build_config()
        self.assertIsNotNone(conf)
        self.assertNotEqual({}, conf)
        return conf


class TestMonAPIDetectionPlugin(unittest.TestCase):
    def setUp(self):
        FakeProcesses.name = 'monasca-api'
        super(TestMonAPIDetectionPlugin, self).setUp()
        with mock.patch.object(mon.MonAPI, '_detect') as mock_detect:
            self._mon_api = mon.MonAPI('foo')
            self.assertTrue(mock_detect.called)

    @mock.patch('monasca_setup.detection.plugins.mon._MonAPIPythonHelper')
    def test_should_use_python_helper_if_api_is_python(self, impl_helper):
        FakeProcesses.cmdLine = [_PYTHON_CMD_API]

        self._mon_api._init_impl_helper = iih = mock.Mock(
            return_value=impl_helper)

        self._detect()

        iih.assert_called_once_with('python')
        self.assertTrue(impl_helper.load_configuration.called_once)
        self.assertTrue(impl_helper.get_bound_port.called_once)

    @mock.patch('monasca_setup.detection.plugins.mon._MonAPIJavaHelper')
    def test_should_use_java_helper_if_api_is_java(self, impl_helper):
        FakeProcesses.cmdLine = [_JAVA_CMD_API]

        self._mon_api._init_impl_helper = iih = mock.Mock(
            return_value=impl_helper)

        self._detect()

        iih.assert_called_once_with('java')
        self.assertTrue(impl_helper.load_configuration.called_once)
        self.assertTrue(impl_helper.get_bound_port.called_once)

    def test_should_detect_java_api_has_config(self):
        app_port = random.randint(1000, 10000)
        admin_port = random.randint(1000, 10000)
        admin_type = 'http'
        if admin_port % 2 != 0:
            admin_type += 's'

        FakeProcesses.cmdLine = [_JAVA_CMD_API]
        FakeProcesses.inetConnections = [FakeInetConnection(app_port)]

        yml_cfg = _JAVA_YML_CFG_BIT_API.format(
            app_port=app_port,
            admin_port=admin_port,
            admin_type=admin_type,
            hibernate_enabled=False
        )

        with mock.patch(
                "__builtin__.open",
                mock.mock_open(read_data=yml_cfg)) as mock_file:
            self._detect()
            mock_file.assert_called_once_with('/etc/monasca/api-config.yml',
                                              'r')

        self.assertTrue(self._mon_api.available)

    @mock.patch('six.moves.configparser.RawConfigParser')
    def test_should_detect_python_api_has_config(self, rcp):
        expected_port = 6666
        actual_port = 6666

        FakeProcesses.cmdLine = [_PYTHON_CMD_API]
        FakeProcesses.inetConnections = [FakeInetConnection(actual_port)]

        # make sure we return the port as we would read from the cfg
        rcp.getint.return_value = expected_port

        # override configuration to make sure we read correct port
        impl_helper = mon._MonAPIPythonHelper()
        impl_helper._paste_config = rcp
        impl_helper.load_configuration = mock.Mock()

        self._mon_api._init_impl_helper = mock.Mock(return_value=impl_helper)

        self._detect()
        self.assertTrue(self._mon_api.available)

    @mock.patch('six.moves.configparser.RawConfigParser')
    def test_should_not_detect_if_port_dont_match(self, rcp):
        expected_port = 6666
        actual_port = 8070

        # assume having python implementation
        FakeProcesses.cmdLine = [_PYTHON_CMD_API]
        FakeProcesses.inetConnections = [FakeInetConnection(actual_port)]

        # make sure we return the port as we would read from the cfg
        rcp.getint.return_value = expected_port

        # override configuration to make sure we read correct port
        impl_helper = mon._MonAPIPythonHelper()
        impl_helper._paste_config = rcp
        impl_helper.load_configuration = mock.Mock()

        self._mon_api._init_impl_helper = mock.Mock(return_value=impl_helper)

        with mock.patch.object(LOG, 'error') as mock_log_error:
            self._detect()
            self.assertFalse(self._mon_api.available)
            mock_log_error.assert_called_with('monasca-api is not listening '
                                              'on port %d. Plugin for '
                                              'monasca-api will not '
                                              'be configured.' % expected_port)

    def test_build_java_config_no_hibernate(self):
        self._run_java_build_config(False)

    def test_build_java_config_with_hibernate(self):
        self._run_java_build_config(True)

    @mock.patch('six.moves.configparser.RawConfigParser')
    def test_build_python_config(self, rcp):
        expected_port = 8070

        FakeProcesses.cmdLine = [_PYTHON_CMD_API]
        FakeProcesses.inetConnections = [FakeInetConnection(expected_port)]

        rcp.getint.return_value = expected_port

        impl_helper = mon._MonAPIPythonHelper()
        impl_helper._paste_config = rcp
        impl_helper.load_configuration = mock.Mock()

        self._mon_api._init_impl_helper = mock.Mock(return_value=impl_helper)

        self._detect()
        conf = self._build_config()

        for key in ('process', ):
            self.assertIn(key, conf)
            bit = conf[key]
            self.assertIsNotNone(bit)
            self.assertNotEqual({}, bit)

    def _run_java_build_config(self, hibernate_enabled):
        FakeProcesses.cmdLine = [_JAVA_CMD_API]
        app_port = random.randint(1000, 10000)
        admin_port = random.randint(1000, 10000)
        admin_type = 'http'
        if admin_port % 2 != 0:
            admin_type += 's'

        FakeProcesses.cmdLine = [_JAVA_CMD_API]
        FakeProcesses.inetConnections = [FakeInetConnection(app_port)]

        # note(trebskit) this is always set to 2
        jvm_metrics_count = 2
        internal_metrics_count = 1
        sql_timers_count = 2

        total_metrics_count = jvm_metrics_count + internal_metrics_count + (
            sql_timers_count if not hibernate_enabled else 0)

        yml_cfg = _JAVA_YML_CFG_BIT_API.format(
            app_port=app_port,
            admin_port=admin_port,
            admin_type=admin_type,
            hibernate_enabled=hibernate_enabled
        )

        with mock.patch(
                "__builtin__.open",
                mock.mock_open(read_data=yml_cfg)) as mf:
            self._detect()
            conf = self._build_config()
            mf.assert_called_once_with('/etc/monasca/api-config.yml',
                                       'r')

        for key in ('http_check', 'http_metrics', 'process'):
            self.assertIn(key, conf)
            bit = conf[key]
            self.assertIsNotNone(bit)
            self.assertNotEqual({}, bit)

        # verify http_check
        http_check_instance = conf['http_check']['instances'][0]
        self.assertFalse(http_check_instance['include_content'])
        self.assertEqual('monitoring-monasca-api healthcheck',
                         http_check_instance['name'])
        self.assertEqual(5, http_check_instance['timeout'])
        self.assertEqual('%s://localhost:%d/healthcheck'
                         % (admin_type, admin_port),
                         http_check_instance['url'])

        # verify http_metrics
        http_metrics_instance = conf['http_metrics']['instances'][0]
        self.assertEqual('monitoring-monasca-api metrics',
                         http_metrics_instance['name'])
        self.assertEqual('%s://localhost:%d/metrics'
                         % (admin_type, admin_port),
                         http_metrics_instance['url'])
        hmi_whitelist = http_metrics_instance['whitelist']
        self.assertIsNotNone(hmi_whitelist)
        self.assertEqual(len(hmi_whitelist), total_metrics_count)

        jvm_metrics_found = 0
        internal_metrics_found = 0
        sql_timers_metrics_found = 0

        for entry in hmi_whitelist:
            name = entry['name']
            if 'jvm' in name:
                jvm_metrics_found += 1
            elif 'metrics.published' in name:
                internal_metrics_found += 1
            elif 'raw-sql.time' in name:
                sql_timers_metrics_found += 1

        self.assertEqual(jvm_metrics_count, jvm_metrics_found)
        self.assertEqual(internal_metrics_count, internal_metrics_found)
        if not hibernate_enabled:
            self.assertEqual(sql_timers_count, sql_timers_count)

    def _build_config(self):
        conf = self._mon_api.build_config()
        self.assertIsNotNone(conf)
        self.assertNotEqual({}, conf)
        return conf

    def _detect(self):
        self._mon_api.available = False

        process_iter = mock.patch.object(psutil, 'process_iter',
                                         return_value=[FakeProcesses()])

        with process_iter as mock_process_iter:
            self._mon_api._detect()
            self.assertTrue(mock_process_iter.called)



import unittest
import logging
import subprocess
from tempfile import mkdtemp
import time
import socket

import pymongo

from tests.common import load_check
from nose.plugins.skip import SkipTest

PORT1 = 37017
PORT2 = 37018
MAX_WAIT = 150


class TestMongo(unittest.TestCase):

    def wait4mongo(self, process, port):
        # Somehow process.communicate() hangs
        out = process.stdout
        loop = 0
        while True:
            l = out.readline()
            if l.find("[initandlisten] waiting for connections on port") > -1:
                break
            else:
                time.sleep(0.1)
                loop += 1
                if loop >= MAX_WAIT:
                    break

    def setUp(self):
        self.agent_config = {
            'version': '0.1',
            'api_key': 'toto'
        }

        # Initialize the check from checks_d
        self.check = load_check('mongo', {'init_config': {}, 'instances': {}}, self.agent_config)

        # Start 2 instances of Mongo in a replica set
        dir1 = mkdtemp()
        dir2 = mkdtemp()
        try:
            self.p1 = subprocess.Popen(["mongod",
                                        "--dbpath",
                                        dir1,
                                        "--port",
                                        str(PORT1),
                                        "--replSet",
                                        "testset/%s:%d" % (socket.gethostname(),
                                                           PORT2),
                                        "--rest"],
                                       executable="mongod",
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)
            # Sleep until mongo comes online
            self.wait4mongo(self.p1, PORT1)
            if self.p1:
                # Set up replication
                c1 = pymongo.Connection('localhost:%s' % PORT1, slave_okay=True)
                self.p2 = subprocess.Popen(["mongod",
                                            "--dbpath",
                                            dir2,
                                            "--port",
                                            str(PORT2),
                                            "--replSet",
                                            "testset/%s:%d" % (socket.gethostname(),
                                                               PORT1),
                                            "--rest"],
                                           executable="mongod",
                                           stdout=subprocess.PIPE,
                                           stderr=subprocess.PIPE)
                self.wait4mongo(self.p2, PORT2)
                # Waiting before all members are online
                time.sleep(15)
                c1.admin.command("replSetInitiate")
                # Sleep for 15s until replication is stable
                time.sleep(30)
                assert pymongo.Connection('localhost:%s' % PORT2)
        except Exception:
            logging.getLogger().exception("Cannot instantiate mongod properly")

    def tearDown(self):
        try:
            if "p1" in dir(self):
                self.p1.terminate()
            if "p2" in dir(self):
                self.p2.terminate()
        except Exception:
            logging.getLogger().exception("Cannot terminate mongod instances")

    def testMongoCheck(self):
        raise SkipTest('Requires MongoDB')
        self.config = {
            'instances': [{
                'server': "mongodb://localhost:%s/test" % PORT1
            },
                {
                'server': "mongodb://localhost:%s/test" % PORT2
            }]
        }

        # Test mongodb with checks_d
        self.check = load_check('mongo', self.config, self.agent_config)

        # Run the check against our running server
        self.check.check(self.config['instances'][0])
        # Sleep for 1 second so the rate interval >=1
        time.sleep(1)
        # Run the check again so we get the rates
        self.check.check(self.config['instances'][0])

        # Metric assertions
        metrics = self.check.get_metrics()
        assert metrics
        self.assertIsInstance(metrics, list)
        self.assertTrue(len(metrics) > 0)

        metric_val_checks = {
            'mongodb.connections.current': lambda x: x >= 1,
            'mongodb.connections.available': lambda x: x >= 1,
            'mongodb.uptime': lambda x: x >= 0,
            'mongodb.mem.resident': lambda x: x > 0,
            'mongodb.mem.virtual': lambda x: x > 0
        }

        for m in metrics:
            metric_name = m[0]
            if metric_name in metric_val_checks:
                self.assertTrue(metric_val_checks[metric_name](m[2]))

        # Run the check against our running server
        self.check.check(self.config['instances'][1])
        # Sleep for 1 second so the rate interval >=1
        time.sleep(1)
        # Run the check again so we get the rates
        self.check.check(self.config['instances'][1])

        # Metric assertions
        metrics = self.check.get_metrics()
        assert metrics
        self.assertIsInstance(metrics, list)
        self.assertTrue(len(metrics) > 0)

        for m in metrics:
            metric_name = m[0]
            if metric_name in metric_val_checks:
                self.assertTrue(metric_val_checks[metric_name](m[2]))

    def testMongoOldConfig(self):
        raise SkipTest('Requires MongoDB')
        self.agent_config1 = {
            'mongodb_server': "mongodb://localhost:%s/test" % PORT1,
            'version': '0.1',
            'api_key': 'toto'
        }
        conf1 = self.check.parse_agent_config(self.agent_config1)
        self.agent_config2 = {
            'mongodb_server': "mongodb://localhost:%s/test" % PORT2,
            'version': '0.1',
            'api_key': 'toto'
        }
        conf2 = self.check.parse_agent_config(self.agent_config2)

        # Test the first mongodb instance
        self.check = load_check('mongo', conf1, self.agent_config1)

        # Run the check against our running server
        self.check.check(conf1['instances'][0])
        # Sleep for 1 second so the rate interval >=1
        time.sleep(1)
        # Run the check again so we get the rates
        self.check.check(conf1['instances'][0])

        # Metric assertions
        metrics = self.check.get_metrics()
        assert metrics
        self.assertIsInstance(metrics, list)
        self.assertTrue(len(metrics) > 0)

        metric_val_checks = {
            'mongodb.connections.current': lambda x: x >= 1,
            'mongodb.connections.available': lambda x: x >= 1,
            'mongodb.uptime': lambda x: x >= 0,
            'mongodb.mem.resident': lambda x: x > 0,
            'mongodb.mem.virtual': lambda x: x > 0
        }

        for m in metrics:
            metric_name = m[0]
            if metric_name in metric_val_checks:
                self.assertTrue(metric_val_checks[metric_name](m[2]))

        # Test the second mongodb instance
        self.check = load_check('mongo', conf2, self.agent_config2)

        # Run the check against our running server
        self.check.check(conf2['instances'][0])
        # Sleep for 1 second so the rate interval >=1
        time.sleep(1)
        # Run the check again so we get the rates
        self.check.check(conf2['instances'][0])

        # Metric assertions
        metrics = self.check.get_metrics()
        assert metrics
        self.assertIsInstance(metrics, list)
        self.assertTrue(len(metrics) > 0)

        for m in metrics:
            metric_name = m[0]
            if metric_name in metric_val_checks:
                self.assertTrue(metric_val_checks[metric_name](m[2]))

if __name__ == '__main__':
    unittest.main()



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP

import monasca_setup.detection


class Directory(monasca_setup.detection.ServicePlugin):

    """Setup configuration to monitor directory size.
       directory_names example:
                    'directory_names': ['/path/to/directory_1',
                                        '/path/to/directory_2',
                                        ...]
       service_name example:
                    'service_name': 'directory-service'
    """

    def __init__(self, template_dir, overwrite=True, args=None):
        service_params = {
            'args': args,
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': '',
            'directory_names': []}
        super(Directory, self).__init__(service_params)



# (C) Copyright 2016 Hewlett Packard Enterprise Development LP

import collections
import mock
import unittest

import monasca_setup
import monasca_setup.agent_config

DEFAULT_HTTP_CHECK_CONFIG = {
    'init_config': None,
    'instances': [{'built_by': 'HttpCheck',
                   'match_pattern': '.*VERSION.*',
                   'url': 'http://127.0.0.1:9200',
                   'name': 'logging',
                   'timeout': '10',
                   'collect_response_time': True,
                   'use_keystone': False,
                   'dimensions': {'service': 'logging'}}]
}

DEFAULT_PROCESS_CHECK_CONFIG = {
    'init_config': None,
    'instances': [{'built_by': 'MonNotification',
                   'detailed': True,
                   'dimensions': {'component': 'monasca-notification'},
                   'exact_match': False,
                   'name': 'monasca-notification',
                   'search_string': ['monasca-notification']
                   }]
}

INPUT_ARGS = collections.namedtuple(
    "InputArgs", ["overwrite", "user", "config_dir", "detection_args",
                  "detection_plugins", "dry_run"])


class TestModifyConfig(unittest.TestCase):
    """ Unit tests for modify_config function in monasca_setup/main.py
        More details are documented in:
        monasca-agent/docs/DeveloperDocs/agent_internals.md
    """
    def setUp(self):
        self._config_data = {}

    def save_config(self, config_dir, plugin_name, user, data):
        self._config_data = data

    @mock.patch('monasca_setup.main.agent_config.save_plugin_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_no_modify_process_check_config(self, mock_read_config,
                                            mock_save_config):
        mock_read_config.return_value = DEFAULT_PROCESS_CHECK_CONFIG
        mock_save_config.side_effect = self.save_config

        # Add a new process check instance
        same_built_by = DEFAULT_PROCESS_CHECK_CONFIG['instances'][0][
            'built_by']
        same_name = DEFAULT_PROCESS_CHECK_CONFIG['instances'][0][
            'name']
        args, detected_config = self._get_mon_api_check_args_and_config(
            same_built_by, same_name)
        self._check_no_change(args, detected_config)

    @mock.patch('monasca_setup.main.agent_config.save_plugin_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_modify_process_check_config(self, mock_read_config,
                                         mock_save_config):
        mock_read_config.return_value = DEFAULT_PROCESS_CHECK_CONFIG
        mock_save_config.side_effect = self.save_config

        # Add a new process check instance
        built_by = 'MonAPI'
        name = 'monasca-api'
        args, detected_config = self._get_mon_api_check_args_and_config(
            built_by, name)
        expected_value = \
            {'init_config': None,
             'instances': [{'built_by': built_by,
                            'detailed': True,
                            'dimensions': {'component': 'monasca-api'},
                            'exact_match': False,
                            'name': name,
                            'search_string': ['monasca-api']
                            },
                           DEFAULT_PROCESS_CHECK_CONFIG['instances'][0]]
             }
        self._check_changes(args, detected_config, expected_value)

    @mock.patch('monasca_setup.main.agent_config.save_plugin_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_no_modify_http_check_config(self, mock_read_config,
                                         mock_save_config):
        mock_read_config.return_value = DEFAULT_HTTP_CHECK_CONFIG
        mock_save_config.side_effect = self.save_config

        # keep url and match_pattern are the same
        same_url = DEFAULT_HTTP_CHECK_CONFIG['instances'][0]['url']
        same_match_pattern = DEFAULT_HTTP_CHECK_CONFIG['instances'][0][
            'match_pattern']
        same_name = DEFAULT_HTTP_CHECK_CONFIG['instances'][0]['name']

        args, detected_config = self. _get_http_check_args_and_config(
            same_url, same_match_pattern, same_name)
        self._check_no_change(args, detected_config)

    @mock.patch('monasca_setup.main.agent_config.save_plugin_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_modify_http_check_config(self, mock_read_config,
                                      mock_save_config):
        mock_read_config.return_value = DEFAULT_HTTP_CHECK_CONFIG
        mock_save_config.side_effect = self.save_config

        # Change protocol and match_pattern
        url = 'https://127.0.0.1:9200'
        match_pattern = '.*OK.*'

        args, detected_config = self. _get_http_check_args_and_config(
            url, match_pattern, 'logging')
        expected_value = {'init_config': None,
                          'instances':
                          [{'built_by': 'HttpCheck',
                            'collect_response_time': True,
                            'url': url,
                            'match_pattern': match_pattern,
                            'name': 'logging',
                            'timeout': '10',
                            'use_keystone': False,
                            'dimensions': {'service': 'logging'}}]
                          }
        self._check_changes(args, detected_config, expected_value)

    @mock.patch('monasca_setup.main.agent_config.save_plugin_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_http_check_endpoint_change(self, mock_read_config,
                                        mock_save_config):
        mock_read_config.return_value = DEFAULT_HTTP_CHECK_CONFIG
        mock_save_config.side_effect = self.save_config

        # Change only protocol
        new_url = 'https://127.0.0.1:9200'
        same_match_pattern = DEFAULT_HTTP_CHECK_CONFIG['instances'][0][
            'match_pattern']
        same_name = DEFAULT_HTTP_CHECK_CONFIG['instances'][0]['name']
        args, detected_config = self. _get_http_check_args_and_config(
            new_url, same_match_pattern, same_name)
        expected_value = {'init_config': None,
                          'instances':
                          [{'built_by': 'HttpCheck',
                            'collect_response_time': True,
                            'url': new_url,
                            'match_pattern': same_match_pattern,
                            'name': same_name,
                            'timeout': '10',
                            'use_keystone': False,
                            'dimensions': {'service': 'logging'}}]
                          }
        self._check_changes(args, detected_config, expected_value)

    @mock.patch('monasca_setup.main.agent_config.save_plugin_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_http_check_only_match_pattern(self, mock_read_config,
                                           mock_save_config):
        mock_read_config.return_value = DEFAULT_HTTP_CHECK_CONFIG
        mock_save_config.side_effect = self.save_config

        # Change only match_pattern
        same_url = DEFAULT_HTTP_CHECK_CONFIG['instances'][0]['url']
        new_match_pattern = '.*TEST.*'
        same_name = DEFAULT_HTTP_CHECK_CONFIG['instances'][0]['name']
        args, detected_config = self. _get_http_check_args_and_config(
            same_url, new_match_pattern, same_name)
        expected_value = {'init_config': None,
                          'instances':
                          [{'built_by': 'HttpCheck',
                            'collect_response_time': True,
                            'url': same_url,
                            'match_pattern': new_match_pattern,
                            'name': same_name,
                            'timeout': '10',
                            'use_keystone': False,
                            'dimensions': {'service': 'logging'}}]
                          }
        self._check_changes(args, detected_config, expected_value)

    @mock.patch('monasca_setup.main.agent_config.save_plugin_config')
    @mock.patch('monasca_setup.main.agent_config.read_plugin_config_from_disk')
    def test_http_check_new_url(self, mock_read_config, mock_save_config):
        mock_read_config.return_value = DEFAULT_HTTP_CHECK_CONFIG
        mock_save_config.side_effect = self.save_config

        # Pass in a new url
        new_url = 'http://192.168.10.6:8070'
        new_match_pattern = '.*TEST.*'
        new_name = new_url
        args, detected_config = self. _get_http_check_args_and_config(
            new_url, new_match_pattern, new_name)
        expected_value = {'init_config': None,
                          'instances':
                          [{'built_by': 'HttpCheck',
                            'collect_response_time': True,
                            'url': new_url,
                            'match_pattern': new_match_pattern,
                            'name': new_name,
                            'timeout': '10',
                            'use_keystone': False,
                            'dimensions': {'service': 'logging'}},
                           DEFAULT_HTTP_CHECK_CONFIG['instances'][0]
                           ]
                          }
        self._check_changes(args, detected_config, expected_value)

    def _check_no_change(self, args, detected_config):
        changes = monasca_setup.main.modify_config(args, detected_config)
        self.assertEqual(changes, False)
        self.assertEqual(self._config_data, {})

    def _check_changes(self, args, detected_config, expected_value):
        changes = monasca_setup.main.modify_config(args, detected_config)
        self.assertEqual(changes, True)
        self.assertEqual(expected_value, self._config_data)

    def _get_mon_api_check_args_and_config(self, built_by, name):
        args = INPUT_ARGS(False, 'mon-agent', '/etc/monasca/agent', None,
                          ['MonAPI'], False)
        detected_config = {
            'process':
                {'instances': [{'built_by': built_by,
                                'detailed': True,
                                'dimensions':  {
                                    'component': name},
                                'exact_match': False,
                                'name': name,
                                'search_string': [name]
                                }],
                 'init_config': None
                 }
        }
        return args, detected_config

    def _get_http_check_args_and_config(self, url, match_pattern, name):
        args = INPUT_ARGS(False, 'mon-agent', '/etc/monasca/agent',
                          'url={0} match_pattern={1} name={2} timeout=10 '
                          'use_keystone=False'.format(
                              url, match_pattern, name),
                          ['HttpCheck'], False)
        detected_config = {
            'http_check':
                {'instances': [{'name': name,
                                'url': url,
                                'built_by': 'HttpCheck',
                                'use_keystone': False,
                                'match_pattern': match_pattern,
                                'collect_response_time': True,
                                'timeout': '10',
                                'dimensions': {'service': 'logging'}
                                }],
                 'init_config': None
                 }
        }
        return args, detected_config



import monasca_setup.detection


class Barbican(monasca_setup.detection.ServicePlugin):

    """Detect Barbican daemons and setup configuration to monitor them."""

    def __init__(self, template_dir, overwrite=True, args=None):
        service_params = {
            'args': args,
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': 'key-manager',
            'process_names': ['barbican-api'],
            'service_api_url': 'http://localhost:9311',
            'search_pattern': '.*v1.*'
        }

        super(Barbican, self).__init__(service_params)



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP

import logging
import os

import yaml

import monasca_setup.agent_config
import monasca_setup.detection

log = logging.getLogger(__name__)


class Zookeeper(monasca_setup.detection.Plugin):

    """Detect Zookeeper daemons and setup configuration to monitor them.

    """

    def _detect(self):
        """Run detection, set self.available True if the service is detected.

        """
        if monasca_setup.detection.find_process_cmdline('org.apache.zookeeper') is not None:
            self.available = True

    def build_config(self):
        """Build the config as a Plugins object and return.

        """
        config = monasca_setup.agent_config.Plugins()
        # First watch the process
        log.info("\tWatching the zookeeper process.")
        config.merge(monasca_setup.detection.watch_process(['org.apache.zookeeper.server'], 'zookeeper',
                                                           exact_match=False))

        log.info("\tEnabling the zookeeper plugin")
        with open(os.path.join(self.template_dir, 'conf.d/zk.yaml.example'), 'r') as zk_template:
            zk_config = yaml.safe_load(zk_template.read())
        config['zk'] = zk_config

        return config

    def dependencies_installed(self):
        # The current plugin just does a simple socket connection to zookeeper and
        # parses the stat command
        return True



import mock
import unittest

from tests.common import load_check


class TestSimpleProcess(unittest.TestCase):
    def setUp(self):
        self.psutil_process_iter_patcher = mock.patch('psutil.process_iter')

        self.mock_process_iter = self.psutil_process_iter_patcher.start()

        process_attrs = {
            'name': 'process_name',
            'pid': 1234,
            'username': 'user',
            'cmdline': '/usr/bin/process_name'
        }
        process = mock.Mock()
        process.as_dict.return_value = process_attrs
        self.mock_process_iter.return_value = [process]

        config = {'init_config': {},
                  'instances': [{'name': 'test',
                                 'search_string': ['process_name'],
                                 'detailed': False}]}
        self.check = load_check('process', config)

    def tearDown(self):
        self.psutil_process_iter_patcher.stop()

    def testPidCount(self):
        self.check.run()
        metrics = self.check.get_metrics()

        self.assertEqual(1, len(metrics))
        self.assertEqual('process.pid_count', metrics[0]['measurement']['name'])


class TestDetailedProcess(unittest.TestCase):
    def setUp(self):
        self.psutil_process_patcher = mock.patch('psutil.Process')
        self.psutil_process_iter_patcher = mock.patch('psutil.process_iter')

        self.mock_process = self.psutil_process_patcher.start()
        self.mock_process_iter = self.psutil_process_iter_patcher.start()

        process_attrs_as_dict = {
            'name': 'process_name',
            'pid': 1234,
            'username': 'user',
            'cmdline': '/usr/bin/process_name',
        }

        process_attrs = {
            'memory_info_ex.return_value': mock.Mock(rss=1048576),
            'num_threads.return_value': 1,
            'num_fds.return_value': 1,
            'cpu_percent.return_value': 1,
            'io_counters.return_value': mock.Mock(**{'read_count': 1,
                                                     'write_count': 1,
                                                     'read_bytes': 1024,
                                                     'write_bytes': 1024})
        }
        process = mock.Mock(**process_attrs)
        process.as_dict.return_value = process_attrs_as_dict
        self.mock_process_iter.return_value = [process]
        self.mock_process.return_value = process

        config = {'init_config': {},
                  'instances': [{'name': 'test',
                                 'search_string': ['process_name'],
                                 'detailed': True}]}
        self.check = load_check('process', config)

    def tearDown(self):
        self.psutil_process_patcher.stop()
        self.psutil_process_iter_patcher.stop()

    def testPidCount(self):
        self.check.run()
        metrics = self.check.get_metrics()

        self.assertGreater(len(metrics), 1)

    def run_check(self):
        self.check.prepare_run()
        self.check.run()
        metrics = self.check.get_metrics()

        measurement_names = [metric['measurement']['name'] for metric in metrics]

        measurement_names.sort()
        return measurement_names

    def testMeasurements(self):
        measurement_names = self.run_check()

        # first run will not have cpu_perc in it
        expected_names = ['process.io.read_count',
                          'process.io.read_kbytes',
                          'process.io.write_count',
                          'process.io.write_kbytes',
                          'process.mem.rss_mbytes',
                          'process.open_file_descriptors',
                          'process.pid_count',
                          'process.thread_count']

        self.assertListEqual(measurement_names, expected_names)

        # run again to get cpu_perc
        expected_names.insert(0, 'process.cpu_perc')
        measurement_names = self.run_check()
        self.assertListEqual(measurement_names, expected_names)



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP

import logging
import platform
import sys

import linux


log = logging.getLogger(__name__)


def detect_init(*args, **kwargs):
    """Detect the service manager running on this box
       args/kwargs match those of service.Service
    :return: The appropriate Service object for this system
    """
    detected_os = platform.system()
    if detected_os == 'Linux':
        supported_linux_flavors = [
            'ubuntu', 'debian',
            'centos linux', 'red hat enterprise linux server',
            'suse linux enterprise server'
        ]
        flavor = platform.linux_distribution()[0]
        if flavor.lower() not in supported_linux_flavors:
            log.warn('{0} is not a support Linux distribution'.format(flavor))
        return detect_linux_init(*args, **kwargs)
    else:
        print("{0} is not currently supported by the Monasca Agent".format(detected_os))
        sys.exit(1)

    # Service enable, includes setup of users/config directories so must be
    # done before configuration


def detect_linux_init(*args, **kwargs):
    """Detect which of the linux inits is running
    :return: Return a valid Linux service manager object
    """
    with open('/proc/1/comm', 'r') as init_proc:
        init = init_proc.readline().strip()
        if init == 'systemd':
            return linux.Systemd(*args, **kwargs)
        else:
            return linux.SysV(*args, **kwargs)



# Copyright 2016 FUJITSU LIMITED
# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os

import requests

from monasca_setup import agent_config
from monasca_setup import detection
from monasca_setup.detection import utils

LOG = logging.getLogger(__name__)

_KIBANA_CFG_FILE = '/opt/kibana/config/kibana.yml'
_API_STATUS = 'api/status'
_METRIC_ALIASES = {
    'heap_total': 'heap_size',
    'requests_per_second': 'req_sec',
    'response_time_avg': 'resp_time_avg',
    'response_time_max': 'resp_time_max'
}


def _to_snake_case(word):
    final = ''
    for item in word:
        if item.isupper():
            final += "_" + item.lower()
        else:
            final += item
    if final[0] == "_":
        final = final[1:]
    return final


def get_metric_name(metric):
    actual_name = _to_snake_case(metric)
    return _METRIC_ALIASES.get(actual_name, actual_name)


class Kibana(detection.Plugin):
    def _detect(self):
        # check process and port

        process_found = utils.find_process_cmdline('kibana') is not None
        has_deps = self.dependencies_installed()

        has_args = self.args is not None
        cfg_file = self._get_config_file() if has_args else _KIBANA_CFG_FILE
        has_config_file = os.path.isfile(cfg_file)

        available = process_found and has_deps and has_config_file

        self.available = available

        if not self.available:
            err_str = 'Plugin for Kibana will not be configured.'
            if not process_found:
                LOG.error('Kibana process has not been found. %s' % err_str)
            elif not has_deps:
                LOG.error('Kibana plugin dependencies are not satisfied. '
                          'Module "pyaml" not found. %s'
                          % err_str)
            elif not has_config_file:
                LOG.error('Kibana plugin cannot find configuration file %s. %s'
                          % (cfg_file, err_str))

    def build_config(self):
        kibana_config = self._get_config_file()

        try:
            (kibana_host,
             kibana_port,
             kibana_protocol) = self._read_config(kibana_config)
        except Exception as ex:
            LOG.error('Failed to read configuration at %s' % kibana_config)
            LOG.exception(ex)
            return

        if kibana_protocol == 'https':
            LOG.error('"https" protocol is currently not supported')
            return None

        config = agent_config.Plugins()

        # retrieve user name and set in config
        # if passed in args (note args are optional)
        if (self.args and 'kibana-user' in self.args and
                self.args['kibana-user']):
            process = detection.watch_process_by_username(
                username=self.args['kibana-user'],
                process_name='kibana',
                service='monitoring',
                component='kibana'
            )
        else:
            process = detection.watch_process(['kibana'],
                                              service='monitoring',
                                              component='kibana',
                                              process_name='kibana')

        config.merge(process)

        kibana_url = '%s://%s:%d' % (
            kibana_protocol,
            kibana_host,
            kibana_port
        )

        if not self._has_metrics_support(kibana_url):
            LOG.warning('Running kibana does not support metrics, skipping...')
            return None
        else:
            metrics = self._get_all_metrics(kibana_url)
            config['kibana'] = {
                'init_config': {
                    'url': '%s/%s' % (kibana_url, _API_STATUS),
                },
                'instances': [
                    {
                        'metrics': metrics
                    }
                ]
            }

        LOG.info('\tWatching the kibana process.')

        return config

    def dependencies_installed(self):
        try:
            import yaml
        except Exception:
            return False
        return True

    def _get_config_file(self):
        if self.args is not None:
            kibana_config = self.args.get('kibana-config', _KIBANA_CFG_FILE)
        else:
            kibana_config = _KIBANA_CFG_FILE
        return kibana_config

    @staticmethod
    def _read_config(kibana_cfg):
        import yaml
        with open(kibana_cfg, 'r') as stream:
            document = yaml.safe_load(stream=stream)

            has_ssl_support = ('server.ssl.cert' in document and
                               'server.ssl.key' in document)

            host = document.get('server.host')
            port = int(document.get('server.port'))
            protocol = 'https' if has_ssl_support else 'http'

            return host, port, protocol

    def _get_all_metrics(self, kibana_url):
        resp = self._get_metrics_request(kibana_url)
        data = resp.json()

        metrics = []
        # do not check plugins, check will go for overall status

        # get metrics
        for metric in data.get('metrics').keys():
            metrics.append(get_metric_name(metric))

        return metrics

    def _has_metrics_support(self, kibana_url):
        resp = self._get_metrics_request(kibana_url, method='HEAD')
        status_code = resp.status_code
        # although Kibana will respond with 400:Bad Request
        # it means that URL is available but simply does
        # not support HEAD request
        # Looks like guys from Kibana just support GET for this url
        return status_code == 400

    def _get_metrics_request(self, url, method='GET'):
        request_url = '%s/%s' % (url, _API_STATUS)
        return requests.request(method=method, url=request_url)



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP

""" Systemd based service
"""
import glob
import logging
import os
import pwd
import subprocess

import service


log = logging.getLogger(__name__)


class LinuxInit(service.Service):
    """Parent class for all Linux based init systems.
    """
    def enable(self):
        """Does user/group directory creation.
        """
        # Create user/group if needed
        try:
            user = pwd.getpwnam(self.username)
        except KeyError:
            subprocess.check_call(['useradd', '-r', self.username])
            user = pwd.getpwnam(self.username)

        # Create dirs
        # todo log dir is hardcoded
        for path in (self.log_dir, self.config_dir, '%s/conf.d' % self.config_dir):
            if not os.path.exists(path):
                os.makedirs(path, 0o755)
                os.chown(path, 0, user.pw_gid)
        # the log dir needs to be writable by the user
        os.chown(self.log_dir, user.pw_uid, user.pw_gid)

    def start(self, restart=True):
        if not self.is_enabled():
            log.error('The service is not enabled')
            return False

    def stop(self):
        if not self.is_enabled():
            log.error('The service is not enabled')
            return True

    def is_enabled(self):
        """Returns True if monasca-agent is setup to start on boot, false otherwise.

        """
        raise NotImplementedError


class Systemd(LinuxInit):
    def enable(self):
        """Sets monasca-agent to start on boot.

            Generally this requires running as super user
        """
        LinuxInit.enable(self)

        # Write the systemd script
        init_path = '/etc/systemd/system/{0}.service'.format(self.name)
        with open(os.path.join(self.template_dir, 'monasca-agent.service.template'), 'r') as template:
            with open(init_path, 'w') as service_script:
                service_script.write(template.read().format(prefix=self.prefix_dir, monasca_user=self.username,
                                                            config_dir=self.config_dir))
        os.chown(init_path, 0, 0)
        os.chmod(init_path, 0o644)

        # Enable the service
        subprocess.check_call(['systemctl', 'daemon-reload'])
        subprocess.check_call(['systemctl', 'enable', '{0}.service'.format(self.name)])
        log.info('Enabled {0} service via systemd'.format(self.name))

    def start(self, restart=True):
        """Starts monasca-agent.

            If the agent is running and restart is True, restart
        """
        LinuxInit.start(self)
        log.info('Starting {0} service via systemd'.format(self.name))
        if restart:
            subprocess.check_call(['systemctl', 'restart', '{0}.service'.format(self.name)])
        else:
            subprocess.check_call(['systemctl', 'start', '{0}.service'.format(self.name)])

        return True

    def stop(self):
        """Stops monasca-agent.
        """
        LinuxInit.stop(self)
        log.info('Stopping {0} service'.format(self.name))
        subprocess.check_call(['systemctl', 'stop', '{0}.service'.format(self.name)])
        return True

    def is_enabled(self):
        """Returns True if monasca-agent is setup to start on boot, false otherwise.
        """
        try:
            subprocess.check_output(['systemctl', 'is-enabled', '{0}.service'.format(self.name)])
        except subprocess.CalledProcessError:
            return False

        return True


class SysV(LinuxInit):

    def __init__(self, prefix_dir, config_dir, log_dir, template_dir, username, name='monasca-agent'):
        """Setup this service with the given init template.

        """
        service.Service.__init__(self, prefix_dir, config_dir, log_dir, template_dir, name, username)
        self.init_script = '/etc/init.d/%s' % self.name
        self.init_template = os.path.join(template_dir, 'monasca-agent.init.template')

    def enable(self):
        """Sets monasca-agent to start on boot.

            Generally this requires running as super user
        """
        LinuxInit.enable(self)
        # Write the init script and enable.
        with open(self.init_template, 'r') as template:
            with open(self.init_script, 'w') as conf:
                conf.write(template.read().format(prefix=self.prefix_dir, monasca_user=self.username,
                                                  config_dir=self.config_dir))
        os.chown(self.init_script, 0, 0)
        os.chmod(self.init_script, 0o755)

        for runlevel in ['2', '3', '4', '5']:
            link_path = '/etc/rc%s.d/S10monasca-agent' % runlevel
            if not os.path.exists(link_path):
                os.symlink(self.init_script, link_path)

        log.info('Enabled {0} service via SysV init script'.format(self.name))

    def start(self, restart=True):
        """Starts monasca-agent.

            If the agent is running and restart is True, restart
        """
        LinuxInit.start(self)

        log.info('Starting {0} service via SysV init script'.format(self.name))
        if restart:
            subprocess.check_call([self.init_script, 'restart'])  # Throws CalledProcessError on error
        else:
            subprocess.check_call([self.init_script, 'start'])  # Throws CalledProcessError on error
        return True

    def stop(self):
        """Stops monasca-agent.

        """
        LinuxInit.stop(self)

        log.info('Stopping {0} service via SysV init script'.format(self.name))
        subprocess.check_call([self.init_script, 'stop'])  # Throws CalledProcessError on error
        return True

    def is_enabled(self):
        """Returns True if monasca-agent is setup to start on boot, false otherwise.

        """
        if not os.path.exists(self.init_script):
            return False

        if len(glob.glob('/etc/rc?.d/S??monasca-agent')) > 0:
            return True
        else:
            return False



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP

import logging
import psutil

log = logging.getLogger(__name__)

import monasca_agent.collector.checks as checks


class Memory(checks.AgentCheck):

    def __init__(self, name, init_config, agent_config):
        super(Memory, self).__init__(name, init_config, agent_config)

    def check(self, instance):
        """Capture memory stats

        """
        dimensions = self._set_dimensions(None, instance)

        mem_info = psutil.virtual_memory()
        swap_info = psutil.swap_memory()

        self.gauge('mem.total_mb',
                   int(mem_info.total / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.free_mb',
                   int(mem_info.free / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.usable_mb',
                   int(mem_info.available / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.used_mb',
                   int(mem_info.used / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.usable_perc',
                   float(100 - mem_info.percent),
                   dimensions=dimensions)
        self.gauge('mem.swap_total_mb',
                   int(swap_info.total / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.swap_used_mb',
                   int(swap_info.used / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.swap_free_mb',
                   int(swap_info.free / 1048576),
                   dimensions=dimensions)
        self.gauge('mem.swap_free_perc',
                   float(100 - swap_info.percent),
                   dimensions=dimensions)

        count = 9
        if hasattr(mem_info, 'buffers') and mem_info.buffers:
            self.gauge('mem.used_buffers',
                       int(mem_info.buffers / 1048576),
                       dimensions=dimensions)
            count += 1

        if hasattr(mem_info, 'cached') and mem_info.cached:
            self.gauge('mem.used_cache',
                       int(mem_info.cached / 1048576),
                       dimensions=dimensions)
            count += 1

        if (hasattr(mem_info, 'buffers') and mem_info.buffers and
           hasattr(mem_info, 'cached') and mem_info.cached):
            self.gauge('mem.used_real_mb',
                       int((mem_info.used - mem_info.buffers - mem_info.cached) / 1048576),
                       dimensions=dimensions)
            count += 1

        if hasattr(mem_info, 'shared') and mem_info.shared:
            self.gauge('mem.used_shared',
                       int(mem_info.shared / 1048576),
                       dimensions=dimensions)
            count += 1

        log.debug('Collected {0} memory metrics'.format(count))



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development Company LP

import logging
import os
import psutil
import re

log = logging.getLogger(__name__)

import monasca_agent.collector.checks as checks


class Disk(checks.AgentCheck):

    def __init__(self, name, init_config, agent_config):
        self._partition_error = set()

        super(Disk, self).__init__(name, init_config, agent_config)

    def _log_once_per_day(self, message):
        if message in self._partition_error:
            return
        self._partition_error.add(message)
        log.exception(message)

    def check(self, instance):
        """Capture disk stats

        """
        dimensions = self._set_dimensions(None, instance)
        rollup_dimensions = dimensions.copy()

        if instance is not None:
            use_mount = instance.get("use_mount", True)
            send_io_stats = instance.get("send_io_stats", True)
            send_rollup_stats = instance.get("send_rollup_stats", False)
            # If we filter devices, get the list.
            device_blacklist_re = self._get_re_exclusions(instance)
            fs_types_to_ignore = self._get_fs_exclusions(instance)
        else:
            use_mount = True
            send_io_stats = True
            send_rollup_stats = False
            device_blacklist_re = None
            fs_types_to_ignore = []

        partitions = psutil.disk_partitions(all=True)
        if send_io_stats:
            disk_stats = psutil.disk_io_counters(perdisk=True)
        disk_count = 0
        total_capacity = 0
        total_used = 0
        for partition in partitions:
            if partition.fstype not in fs_types_to_ignore \
                and (not device_blacklist_re
                     or not device_blacklist_re.match(partition.device)):
                    try:
                        device_name = self._get_device_name(partition.device)
                        disk_usage = psutil.disk_usage(partition.mountpoint)
                        total_capacity += disk_usage.total
                        total_used += disk_usage.used
                        st = os.statvfs(partition.mountpoint)
                    except Exception as ex:
                        exception_name = ex.__class__.__name__
                        self._log_once_per_day('Unable to access partition {} '
                                               'with error: {}'.format(partition,
                                                                       exception_name))
                        continue

                    if use_mount:
                        dimensions.update({'mount_point': partition.mountpoint})
                    self.gauge("disk.space_used_perc",
                               disk_usage.percent,
                               device_name=device_name,
                               dimensions=dimensions)
                    disk_count += 1
                    if st.f_files > 0:
                        self.gauge("disk.inode_used_perc",
                                   round((float(st.f_files - st.f_ffree) / st.f_files) * 100, 2),
                                   device_name=device_name,
                                   dimensions=dimensions)
                        disk_count += 1

                    log.debug('Collected {0} disk usage metrics for partition {1}'.format(disk_count, partition.mountpoint))
                    disk_count = 0
                    if send_io_stats:
                        try:
                            stats = disk_stats[device_name]
                            self.rate("io.read_req_sec", round(float(stats.read_count), 2), device_name=device_name, dimensions=dimensions)
                            self.rate("io.write_req_sec", round(float(stats.write_count), 2), device_name=device_name, dimensions=dimensions)
                            self.rate("io.read_kbytes_sec", round(float(stats.read_bytes / 1024), 2), device_name=device_name, dimensions=dimensions)
                            self.rate("io.write_kbytes_sec", round(float(stats.write_bytes / 1024), 2), device_name=device_name, dimensions=dimensions)
                            self.rate("io.read_time_sec", round(float(stats.read_time / 1000), 2), device_name=device_name, dimensions=dimensions)
                            self.rate("io.write_time_sec", round(float(stats.write_time / 1000), 2), device_name=device_name, dimensions=dimensions)

                            log.debug('Collected 6 disk I/O metrics for partition {0}'.format(partition.mountpoint))
                        except KeyError:
                            log.debug('No Disk I/O metrics available for {0}...Skipping'.format(device_name))

        if send_rollup_stats:
            self.gauge("disk.total_space_mb",
                       total_capacity / 1048576,
                       dimensions=rollup_dimensions)
            self.gauge("disk.total_used_space_mb",
                       total_used / 1048576,
                       dimensions=rollup_dimensions)
            log.debug('Collected 2 rolled-up disk usage metrics')

    def _get_re_exclusions(self, instance):
        """Parse device blacklist regular expression"""
        filter = None
        try:
            filter_device_re = instance.get('device_blacklist_re', None)
            if filter_device_re:
                filter = re.compile(filter_device_re)
        except re.error:
            log.error('Error processing regular expression {0}'.format(filter_device_re))

        return filter

    def _get_fs_exclusions(self, instance):
        """parse comma separated file system types to ignore list"""
        file_system_list = []
        try:
            file_systems = instance.get('ignore_filesystem_types', None)
            if file_systems:
                # Parse file system types
                file_system_list.extend([x.strip() for x in file_systems.split(',')])
        except ValueError:
            log.info("Unable to process ignore_filesystem_types.")

        return file_system_list

    def _get_device_name(self, device):
        start = device.rfind("/")
        if start > -1:
            device_name = device[start + 1:]
        else:
            device_name = device

        return device_name
